@@ -3,18 +3,18 @@
  * xact.c
  *	  top level transaction system support routines
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/access/transam/xact.c,v 1.135.2.3 2006/05/21 20:07:11 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/access/transam/xact.c,v 1.156.2.1 2004/08/11 04:08:00 tgl Exp $
  *
  * NOTES
  *		Transaction aborts can now occur two ways:
  *
- *		1)	system dies from some internal cause  (Assert, etc..)
- *		2)	user types abort
+ *		1)	system dies from some internal cause  (syntax error, etc..)
+ *		2)	user types ABORT
  *
  *		These two cases used to be treated identically, but now
  *		we need to distinguish them.  Why?	consider the following
@@ -30,8 +30,8 @@
  *		In case 1, we want to abort the transaction and return to the
  *		default state.	In case 2, there may be more commands coming
  *		our way which are part of the same transaction block and we have
- *		to ignore these commands until we see an END transaction.
- *		(or an ABORT! --djm)
+ *		to ignore these commands until we see a COMMIT transaction or
+ *		ROLLBACK.
  *
  *		Internal aborts are now handled by AbortTransactionBlock(), just as
  *		they always have been, and user aborts are now handled by
@@ -49,17 +49,9 @@
  *		* CleanupTransaction() executes when we finally see a user COMMIT
  *		  or ROLLBACK command; it cleans things up and gets us out of
  *		  the transaction internally.  In particular, we mustn't destroy
- *		  TransactionCommandContext until this point.
+ *		  TopTransactionContext until this point.
  *
  *	 NOTES
- *		This file is an attempt at a redesign of the upper layer
- *		of the V1 transaction system which was too poorly thought
- *		out to describe.  This new system hopes to be both simpler
- *		in design, simpler to extend and needs to contain added
- *		functionality to solve problems beyond the scope of the V1
- *		system.  (In particuler, communication of transaction
- *		information between parallel backends has to be supported)
- *
  *		The essential aspects of the transaction system are:
  *
  *				o  transaction id generation
@@ -69,7 +61,7 @@
  *				o  lock cleanup
  *
  *		Hence, the functional division of the transaction code is
- *		based on what of the above things need to be done during
+ *		based on which of the above things need to be done during
  *		a start/commit/abort transaction.  For instance, the
  *		routine AtCommit_Memory() takes care of all the memory
  *		cleanup stuff done at commit time.
@@ -99,17 +91,17 @@
  *				CommitTransactionBlock
  *				AbortTransactionBlock
  *
- *		These are invoked only in responce to a user "BEGIN", "END",
- *		or "ABORT" command.  The tricky part about these functions
+ *		These are invoked only in response to a user "BEGIN WORK", "COMMIT",
+ *		or "ROLLBACK" command.	The tricky part about these functions
  *		is that they are called within the postgres main loop, in between
  *		the StartTransactionCommand() and CommitTransactionCommand().
  *
  *		For example, consider the following sequence of user commands:
  *
  *		1)		begin
- *		2)		retrieve (foo.all)
- *		3)		append foo (bar = baz)
- *		4)		end
+ *		2)		select * from foo
+ *		3)		insert into foo (bar = baz)
+ *		4)		commit
  *
  *		in the main processing loop, this results in the following
  *		transaction sequence:
@@ -120,15 +112,15 @@
  *			\	CommitTransactionCommand();
  *
  *			/	StartTransactionCommand();
- *		2) <	ProcessQuery();					<< retrieve (foo.all)
+ *		2) <	ProcessQuery();					<< select * from foo
  *			\	CommitTransactionCommand();
  *
  *			/	StartTransactionCommand();
- *		3) <	ProcessQuery();					<< append foo (bar = baz)
+ *		3) <	ProcessQuery();					<< insert into foo (bar = baz)
  *			\	CommitTransactionCommand();
  *
  *			/	StartTransactionCommand();
- *		4) /	ProcessUtility();				<< end
+ *		4) /	ProcessUtility();				<< commit
  *		   \		CommitTransactionBlock();
  *			\	CommitTransactionCommand();
  *
@@ -139,19 +131,14 @@
  *		outside these calls they need to do normal start/commit
  *		processing.
  *
- *		Furthermore, suppose the "retrieve (foo.all)" caused an abort
+ *		Furthermore, suppose the "select * from foo" caused an abort
  *		condition.	We would then want to abort the transaction and
- *		ignore all subsequent commands up to the "end".
+ *		ignore all subsequent commands up to the "commit".
  *		-cim 3/23/90
  *
  *-------------------------------------------------------------------------
  */
 
-/*
- * Large object clean up added in CommitTransaction() to prevent buffer leaks.
- * [PA, 7/17/98]
- * [PA] is Pascal Andr <andre@via.ecp.fr>
- */
 #include "postgres.h"
 
 #include <unistd.h>
@@ -166,6 +153,7 @@
 #include "catalog/index.h"
 #include "catalog/namespace.h"
 #include "commands/async.h"
+#include "commands/tablecmds.h"
 #include "commands/trigger.h"
 #include "commands/user.h"
 #include "executor/spi.h"
@@ -198,14 +186,14 @@ static void AtCommit_Memory(void);
 static void AtStart_Cache(void);
 static void AtStart_Locks(void);
 static void AtStart_Memory(void);
+static void CallEOXactCallbacks(bool isCommit);
 static void CleanupTransaction(void);
 static void CommitTransaction(void);
 static void RecordTransactionAbort(void);
 static void StartTransaction(void);
 
-/* ----------------
- *		global variables holding the current transaction state.
- * ----------------
+/*
+ *	global variables holding the current transaction state.
  */
 static TransactionStateData CurrentTransactionStateData = {
 	0,							/* transaction id */
@@ -213,47 +201,59 @@ static TransactionStateData CurrentTransactionStateData = {
 	0,							/* scan command id */
 	0x0,						/* start time */
 	TRANS_DEFAULT,				/* transaction state */
-	TBLOCK_DEFAULT				/* transaction block state */
+	TBLOCK_DEFAULT				/* transaction block state from the client
+								 * perspective */
 };
 
-TransactionState CurrentTransactionState = &CurrentTransactionStateData;
+static TransactionState CurrentTransactionState = &CurrentTransactionStateData;
 
 /*
- * User-tweakable parameters
+ *	User-tweakable parameters
  */
 int			DefaultXactIsoLevel = XACT_READ_COMMITTED;
 int			XactIsoLevel;
 
-bool		autocommit = true;
+bool		DefaultXactReadOnly = false;
+bool		XactReadOnly;
 
 int			CommitDelay = 0;	/* precommit delay in microseconds */
 int			CommitSiblings = 5; /* number of concurrent xacts needed to
 								 * sleep */
 
 
-static bool suppressChain = false;
+/*
+ * List of add-on end-of-xact callbacks
+ */
+typedef struct EOXactCallbackItem
+{
+	struct EOXactCallbackItem *next;
+	EOXactCallback callback;
+	void	   *arg;
+} EOXactCallbackItem;
+
+static EOXactCallbackItem *EOXact_callbacks = NULL;
 
 static void (*_RollbackFunc) (void *) = NULL;
 static void *_RollbackData = NULL;
 
 
 /* ----------------------------------------------------------------
- *					 transaction state accessors
+ *	transaction state accessors
  * ----------------------------------------------------------------
  */
 
 #ifdef NOT_USED
 
 /* --------------------------------
- *		TransactionFlushEnabled()
- *		SetTransactionFlushEnabled()
+ *	TransactionFlushEnabled()
+ *	SetTransactionFlushEnabled()
  *
- *		These are used to test and set the "TransactionFlushState"
- *		varable.  If this variable is true (the default), then
- *		the system will flush all dirty buffers to disk at the end
- *		of each transaction.   If false then we are assuming the
- *		buffer pool resides in stable main memory, in which case we
- *		only do writes as necessary.
+ *	These are used to test and set the "TransactionFlushState"
+ *	variable.  If this variable is true (the default), then
+ *	the system will flush all dirty buffers to disk at the end
+ *	of each transaction.   If false then we are assuming the
+ *	buffer pool resides in stable main memory, in which case we
+ *	only do writes as necessary.
  * --------------------------------
  */
 static int	TransactionFlushState = 1;
@@ -272,12 +272,11 @@ SetTransactionFlushEnabled(bool state)
 #endif
 
 
-/* --------------------------------
- *		IsTransactionState
+/*
+ *	IsTransactionState
  *
- *		This returns true if we are currently running a query
- *		within an executing transaction.
- * --------------------------------
+ *	This returns true if we are currently running a query
+ *	within an executing transaction.
  */
 bool
 IsTransactionState(void)
@@ -304,12 +303,11 @@ IsTransactionState(void)
 	return false;
 }
 
-/* --------------------------------
- *		IsAbortedTransactionBlockState
+/*
+ *	IsAbortedTransactionBlockState
  *
- *		This returns true if we are currently running a query
- *		within an aborted transaction block.
- * --------------------------------
+ *	This returns true if we are currently running a query
+ *	within an aborted transaction block.
  */
 bool
 IsAbortedTransactionBlockState(void)
@@ -323,9 +321,8 @@ IsAbortedTransactionBlockState(void)
 }
 
 
-/* --------------------------------
- *		GetCurrentTransactionId
- * --------------------------------
+/*
+ *	GetCurrentTransactionId
  */
 TransactionId
 GetCurrentTransactionId(void)
@@ -336,9 +333,8 @@ GetCurrentTransactionId(void)
 }
 
 
-/* --------------------------------
- *		GetCurrentCommandId
- * --------------------------------
+/*
+ *	GetCurrentCommandId
  */
 CommandId
 GetCurrentCommandId(void)
@@ -349,9 +345,8 @@ GetCurrentCommandId(void)
 }
 
 
-/* --------------------------------
- *		GetCurrentTransactionStartTime
- * --------------------------------
+/*
+ *	GetCurrentTransactionStartTime
  */
 AbsoluteTime
 GetCurrentTransactionStartTime(void)
@@ -362,9 +357,8 @@ GetCurrentTransactionStartTime(void)
 }
 
 
-/* --------------------------------
- *		GetCurrentTransactionStartTimeUsec
- * --------------------------------
+/*
+ *	GetCurrentTransactionStartTimeUsec
  */
 AbsoluteTime
 GetCurrentTransactionStartTimeUsec(int *msec)
@@ -377,14 +371,13 @@ GetCurrentTransactionStartTimeUsec(int *msec)
 }
 
 
-/* --------------------------------
- *		TransactionIdIsCurrentTransactionId
+/*
+ *	TransactionIdIsCurrentTransactionId
  *
- * During bootstrap, we cheat and say "it's not my transaction ID" even though
- * it is.  Along with transam.c's cheat to say that the bootstrap XID is
- * already committed, this causes the tqual.c routines to see previously
- * inserted tuples as committed, which is what we need during bootstrap.
- * --------------------------------
+ *	During bootstrap, we cheat and say "it's not my transaction ID" even though
+ *	it is.	Along with transam.c's cheat to say that the bootstrap XID is
+ *	already committed, this causes the tqual.c routines to see previously
+ *	inserted tuples as committed, which is what we need during bootstrap.
  */
 bool
 TransactionIdIsCurrentTransactionId(TransactionId xid)
@@ -401,9 +394,8 @@ TransactionIdIsCurrentTransactionId(TransactionId xid)
 }
 
 
-/* --------------------------------
- *		CommandIdIsCurrentCommandId
- * --------------------------------
+/*
+ *	CommandIdIsCurrentCommandId
  */
 bool
 CommandIdIsCurrentCommandId(CommandId cid)
@@ -414,9 +406,8 @@ CommandIdIsCurrentCommandId(CommandId cid)
 }
 
 
-/* --------------------------------
- *		CommandCounterIncrement
- * --------------------------------
+/*
+ *	CommandCounterIncrement
  */
 void
 CommandCounterIncrement(void)
@@ -425,7 +416,9 @@ CommandCounterIncrement(void)
 
 	s->commandId += 1;
 	if (s->commandId == FirstCommandId) /* check for overflow */
-		elog(ERROR, "You may only have 2^32-1 commands per transaction");
+		ereport(ERROR,
+				(errcode(ERRCODE_PROGRAM_LIMIT_EXCEEDED),
+				 errmsg("cannot have more than 2^32-1 commands in a transaction")));
 
 	/* Propagate new command ID into query snapshots, if set */
 	if (QuerySnapshot)
@@ -447,9 +440,8 @@ CommandCounterIncrement(void)
  * ----------------------------------------------------------------
  */
 
-/* --------------------------------
- *		AtStart_Cache
- * --------------------------------
+/*
+ *	AtStart_Cache
  */
 static void
 AtStart_Cache(void)
@@ -457,9 +449,8 @@ AtStart_Cache(void)
 	AcceptInvalidationMessages();
 }
 
-/* --------------------------------
+/*
  *		AtStart_Locks
- * --------------------------------
  */
 static void
 AtStart_Locks(void)
@@ -472,21 +463,19 @@ AtStart_Locks(void)
 	 */
 }
 
-/* --------------------------------
- *		AtStart_Memory
- * --------------------------------
+/*
+ *	AtStart_Memory
  */
 static void
 AtStart_Memory(void)
 {
 	/*
-	 * We shouldn't have any transaction contexts already.
+	 * We shouldn't have a transaction context already.
 	 */
 	Assert(TopTransactionContext == NULL);
-	Assert(TransactionCommandContext == NULL);
 
 	/*
-	 * Create a toplevel context for the transaction.
+	 * Create a toplevel context for the transaction, and make it active.
 	 */
 	TopTransactionContext =
 		AllocSetContextCreate(TopMemoryContext,
@@ -495,16 +484,7 @@ AtStart_Memory(void)
 							  ALLOCSET_DEFAULT_INITSIZE,
 							  ALLOCSET_DEFAULT_MAXSIZE);
 
-	/*
-	 * Create a statement-level context and make it active.
-	 */
-	TransactionCommandContext =
-		AllocSetContextCreate(TopTransactionContext,
-							  "TransactionCommandContext",
-							  ALLOCSET_DEFAULT_MINSIZE,
-							  ALLOCSET_DEFAULT_INITSIZE,
-							  ALLOCSET_DEFAULT_MAXSIZE);
-	MemoryContextSwitchTo(TransactionCommandContext);
+	MemoryContextSwitchTo(TopTransactionContext);
 }
 
 
@@ -514,7 +494,7 @@ AtStart_Memory(void)
  */
 
 /*
- *		RecordTransactionCommit
+ *	RecordTransactionCommit
  */
 void
 RecordTransactionCommit(void)
@@ -615,13 +595,13 @@ RecordTransactionCommit(void)
 
 		/*
 		 * We must mark the transaction committed in clog if its XID
-		 * appears either in permanent rels or in local temporary rels.
-		 * We test this by seeing if we made transaction-controlled
-		 * entries *OR* local-rel tuple updates.  Note that if we made
-		 * only the latter, we have not emitted an XLOG record for our
-		 * commit, and so in the event of a crash the clog update might be
-		 * lost.  This is okay because no one else will ever care whether
-		 * we committed.
+		 * appears either in permanent rels or in local temporary rels. We
+		 * test this by seeing if we made transaction-controlled entries
+		 * *OR* local-rel tuple updates.  Note that if we made only the
+		 * latter, we have not emitted an XLOG record for our commit, and
+		 * so in the event of a crash the clog update might be lost.  This
+		 * is okay because no one else will ever care whether we
+		 * committed.
 		 */
 		if (MyLastRecPtr.xrecoff != 0 || MyXactMadeTempRelUpdate)
 			TransactionIdCommit(xid);
@@ -643,9 +623,8 @@ RecordTransactionCommit(void)
 }
 
 
-/* --------------------------------
- *		AtCommit_Cache
- * --------------------------------
+/*
+ *	AtCommit_Cache
  */
 static void
 AtCommit_Cache(void)
@@ -661,9 +640,8 @@ AtCommit_Cache(void)
 	AtEOXactInvalidationMessages(true);
 }
 
-/* --------------------------------
- *		AtCommit_LocalCache
- * --------------------------------
+/*
+ *	AtCommit_LocalCache
  */
 static void
 AtCommit_LocalCache(void)
@@ -674,9 +652,8 @@ AtCommit_LocalCache(void)
 	CommandEndInvalidationMessages(true);
 }
 
-/* --------------------------------
- *		AtCommit_Locks
- * --------------------------------
+/*
+ *	AtCommit_Locks
  */
 static void
 AtCommit_Locks(void)
@@ -689,9 +666,8 @@ AtCommit_Locks(void)
 	ProcReleaseLocks(true);
 }
 
-/* --------------------------------
- *		AtCommit_Memory
- * --------------------------------
+/*
+ *	AtCommit_Memory
  */
 static void
 AtCommit_Memory(void)
@@ -709,7 +685,6 @@ AtCommit_Memory(void)
 	Assert(TopTransactionContext != NULL);
 	MemoryContextDelete(TopTransactionContext);
 	TopTransactionContext = NULL;
-	TransactionCommandContext = NULL;
 }
 
 /* ----------------------------------------------------------------
@@ -718,7 +693,7 @@ AtCommit_Memory(void)
  */
 
 /*
- *		RecordTransactionAbort
+ *	RecordTransactionAbort
  */
 static void
 RecordTransactionAbort(void)
@@ -737,8 +712,7 @@ RecordTransactionAbort(void)
 		 * RecordTransactionCommit ...
 		 */
 		if (TransactionIdDidCommit(xid))
-			elog(PANIC, "RecordTransactionAbort: xact %u already committed",
-				 xid);
+			elog(PANIC, "cannot abort transaction %u, it was already committed", xid);
 
 		START_CRIT_SECTION();
 
@@ -788,9 +762,8 @@ RecordTransactionAbort(void)
 	MyProc->logRec.xrecoff = 0;
 }
 
-/* --------------------------------
- *		AtAbort_Cache
- * --------------------------------
+/*
+ *	AtAbort_Cache
  */
 static void
 AtAbort_Cache(void)
@@ -799,9 +772,8 @@ AtAbort_Cache(void)
 	AtEOXactInvalidationMessages(false);
 }
 
-/* --------------------------------
- *		AtAbort_Locks
- * --------------------------------
+/*
+ *	AtAbort_Locks
  */
 static void
 AtAbort_Locks(void)
@@ -815,28 +787,26 @@ AtAbort_Locks(void)
 }
 
 
-/* --------------------------------
- *		AtAbort_Memory
- * --------------------------------
+/*
+ *	AtAbort_Memory
  */
 static void
 AtAbort_Memory(void)
 {
 	/*
 	 * Make sure we are in a valid context (not a child of
-	 * TransactionCommandContext...).  Note that it is possible for this
-	 * code to be called when we aren't in a transaction at all; go
-	 * directly to TopMemoryContext in that case.
+	 * TopTransactionContext...).  Note that it is possible for this code
+	 * to be called when we aren't in a transaction at all; go directly to
+	 * TopMemoryContext in that case.
 	 */
-	if (TransactionCommandContext != NULL)
+	if (TopTransactionContext != NULL)
 	{
-		MemoryContextSwitchTo(TransactionCommandContext);
+		MemoryContextSwitchTo(TopTransactionContext);
 
 		/*
-		 * We do not want to destroy transaction contexts yet, but it
-		 * should be OK to delete any command-local memory.
+		 * We do not want to destroy the transaction's global state yet,
+		 * so we can't free any memory here.
 		 */
-		MemoryContextResetAndDeleteChildren(TransactionCommandContext);
 	}
 	else
 		MemoryContextSwitchTo(TopMemoryContext);
@@ -848,9 +818,8 @@ AtAbort_Memory(void)
  * ----------------------------------------------------------------
  */
 
-/* --------------------------------
- *		AtCleanup_Memory
- * --------------------------------
+/*
+ *	AtCleanup_Memory
  */
 static void
 AtCleanup_Memory(void)
@@ -868,7 +837,6 @@ AtCleanup_Memory(void)
 	if (TopTransactionContext != NULL)
 		MemoryContextDelete(TopTransactionContext);
 	TopTransactionContext = NULL;
-	TransactionCommandContext = NULL;
 }
 
 
@@ -877,29 +845,19 @@ AtCleanup_Memory(void)
  * ----------------------------------------------------------------
  */
 
-/* --------------------------------
- *		StartTransaction
- *
- * --------------------------------
+/*
+ *	StartTransaction
  */
 static void
 StartTransaction(void)
 {
 	TransactionState s = CurrentTransactionState;
 
-	FreeXactSnapshot();
-	XactIsoLevel = DefaultXactIsoLevel;
-
 	/*
-	 * Check the current transaction state.  If the transaction system is
-	 * switched off, or if we're already in a transaction, do nothing.
-	 * We're already in a transaction when the monitor sends a null
-	 * command to the backend to flush the comm channel.  This is a hacky
-	 * fix to a communications problem, and we keep having to deal with it
-	 * here.  We should fix the comm channel code.	mao 080891
+	 * check the current transaction state
 	 */
-	if (s->state == TRANS_INPROGRESS)
-		return;
+	if (s->state != TRANS_DEFAULT)
+		elog(WARNING, "StartTransaction and not in default state");
 
 	/*
 	 * set the current transaction state information appropriately during
@@ -907,7 +865,12 @@ StartTransaction(void)
 	 */
 	s->state = TRANS_START;
 
-	SetReindexProcessing(false);
+	/*
+	 * Make sure we've freed any old snapshot, and reset xact state variables
+	 */
+	FreeXactSnapshot();
+	XactIsoLevel = DefaultXactIsoLevel;
+	XactReadOnly = DefaultXactReadOnly;
 
 	/*
 	 * generate a new transaction id
@@ -942,22 +905,8 @@ StartTransaction(void)
 
 }
 
-#ifdef NOT_USED
-/* ---------------
- * Tell me if we are currently in progress
- * ---------------
- */
-bool
-CurrentXactInProgress(void)
-{
-	return CurrentTransactionState->state == TRANS_INPROGRESS;
-}
-#endif
-
-/* --------------------------------
- *		CommitTransaction
- *
- * --------------------------------
+/*
+ *	CommitTransaction
  */
 static void
 CommitTransaction(void)
@@ -977,6 +926,12 @@ CommitTransaction(void)
 	 */
 	DeferredTriggerEndXact();
 
+	/*
+	 * Similarly, let ON COMMIT management do its thing before we start to
+	 * commit.
+	 */
+	PreCommit_on_commit_actions();
+
 	/* Prevent cancel/die interrupt while cleaning up */
 	HOLD_INTERRUPTS();
 
@@ -991,7 +946,7 @@ CommitTransaction(void)
 	 * access, and in fact could still cause an error...)
 	 */
 
-	AtEOXact_portals();
+	AtCommit_Portals();
 
 	/* handle commit for large objects [ PA, 7/17/98 ] */
 	/* XXX probably this does not belong here */
@@ -1034,10 +989,10 @@ CommitTransaction(void)
 	 * noncritical resource releasing.
 	 *
 	 * The ordering of operations is not entirely random.  The idea is:
-	 * release resources visible to other backends (eg, files, buffer pins);
-	 * then release locks; then release backend-local resources.  We want
-	 * to release locks at the point where any backend waiting for us will
-	 * see our transaction as being fully cleaned up.
+	 * release resources visible to other backends (eg, files, buffer
+	 * pins); then release locks; then release backend-local resources. We
+	 * want to release locks at the point where any backend waiting for us
+	 * will see our transaction as being fully cleaned up.
 	 */
 
 	smgrDoPendingDeletes(true);
@@ -1047,12 +1002,14 @@ CommitTransaction(void)
 
 	AtCommit_Locks();
 
+	CallEOXactCallbacks(true);
 	AtEOXact_GUC(true);
 	AtEOXact_SPI();
 	AtEOXact_gist();
 	AtEOXact_hash();
 	AtEOXact_nbtree();
 	AtEOXact_rtree();
+	AtEOXact_on_commit_actions(true);
 	AtEOXact_Namespace(true);
 	AtEOXact_CatCache(true);
 	AtEOXact_Files();
@@ -1068,10 +1025,8 @@ CommitTransaction(void)
 	RESUME_INTERRUPTS();
 }
 
-/* --------------------------------
- *		AbortTransaction
- *
- * --------------------------------
+/*
+ *	AbortTransaction
  */
 static void
 AbortTransaction(void)
@@ -1101,14 +1056,9 @@ AbortTransaction(void)
 
 	/*
 	 * check the current transaction state
-	 *
-	 * reduced to DEBUG2 because this is expected when rejecting an
-	 * invalidly-encoded query outside a transaction block.  PG 8.0
-	 * and up fix it better, but it's not worth back-porting those
-	 * changes to 7.3.
 	 */
 	if (s->state != TRANS_INPROGRESS)
-		elog(DEBUG2, "AbortTransaction and not in in-progress state");
+		elog(WARNING, "AbortTransaction and not in in-progress state");
 
 	/*
 	 * set the current transaction state information appropriately during
@@ -1128,7 +1078,7 @@ AbortTransaction(void)
 	 * do abort processing
 	 */
 	DeferredTriggerAbortXact();
-	AtEOXact_portals();
+	AtAbort_Portals();
 	lo_commit(false);			/* 'false' means it's abort */
 	AtAbort_Notify();
 	AtEOXact_UpdatePasswordFile(false);
@@ -1151,7 +1101,7 @@ AbortTransaction(void)
 	}
 
 	/*
-	 * Post-abort cleanup.  See notes in CommitTransaction() concerning
+	 * Post-abort cleanup.	See notes in CommitTransaction() concerning
 	 * ordering.
 	 */
 
@@ -1162,15 +1112,18 @@ AbortTransaction(void)
 
 	AtAbort_Locks();
 
+	CallEOXactCallbacks(false);
 	AtEOXact_GUC(false);
 	AtEOXact_SPI();
 	AtEOXact_gist();
 	AtEOXact_hash();
 	AtEOXact_nbtree();
 	AtEOXact_rtree();
+	AtEOXact_on_commit_actions(false);
 	AtEOXact_Namespace(false);
 	AtEOXact_CatCache(false);
 	AtEOXact_Files();
+	SetReindexProcessing(InvalidOid, InvalidOid);
 	pgstat_count_xact_rollback();
 
 	/*
@@ -1179,10 +1132,8 @@ AbortTransaction(void)
 	RESUME_INTERRUPTS();
 }
 
-/* --------------------------------
- *		CleanupTransaction
- *
- * --------------------------------
+/*
+ *	CleanupTransaction
  */
 static void
 CleanupTransaction(void)
@@ -1198,7 +1149,8 @@ CleanupTransaction(void)
 	/*
 	 * do abort cleanup processing
 	 */
-	AtCleanup_Memory();
+	AtCleanup_Portals();		/* now safe to release portal memory */
+	AtCleanup_Memory();			/* and transaction memory */
 
 	/*
 	 * done with abort processing, set current transaction state back to
@@ -1207,26 +1159,14 @@ CleanupTransaction(void)
 	s->state = TRANS_DEFAULT;
 }
 
-/* --------------------------------
- *		StartTransactionCommand
- *
- * preventChain, if true, forces autocommit behavior at the next
- * CommitTransactionCommand call.
- * --------------------------------
+/*
+ *	StartTransactionCommand
  */
 void
-StartTransactionCommand(bool preventChain)
+StartTransactionCommand(void)
 {
 	TransactionState s = CurrentTransactionState;
 
-	/*
-	 * Remember if caller wants to prevent autocommit-off chaining. This
-	 * is only allowed if not already in a transaction block.
-	 */
-	suppressChain = preventChain;
-	if (preventChain && s->blockState != TBLOCK_DEFAULT)
-		elog(ERROR, "StartTransactionCommand: can't prevent chain");
-
 	switch (s->blockState)
 	{
 			/*
@@ -1293,55 +1233,29 @@ StartTransactionCommand(bool preventChain)
 	}
 
 	/*
-	 * We must switch to TransactionCommandContext before returning. This
-	 * is already done if we called StartTransaction, otherwise not.
+	 * We must switch to TopTransactionContext before returning. This is
+	 * already done if we called StartTransaction, otherwise not.
 	 */
-	Assert(TransactionCommandContext != NULL);
-	MemoryContextSwitchTo(TransactionCommandContext);
+	Assert(TopTransactionContext != NULL);
+	MemoryContextSwitchTo(TopTransactionContext);
 }
 
-/* --------------------------------
- *		CommitTransactionCommand
- *
- * forceCommit = true forces autocommit behavior even when autocommit is off.
- * --------------------------------
+/*
+ *	CommitTransactionCommand
  */
 void
-CommitTransactionCommand(bool forceCommit)
+CommitTransactionCommand(void)
 {
 	TransactionState s = CurrentTransactionState;
 
 	switch (s->blockState)
 	{
 			/*
-			 * If we aren't in a transaction block, and we are doing
-			 * autocommit, just do our usual transaction commit.  But if
-			 * we aren't doing autocommit, start a transaction block
-			 * automatically by switching to INPROGRESS state.	(We handle
-			 * this choice here, and not earlier, so that an explicit
-			 * BEGIN issued in autocommit-off mode won't issue strange
-			 * warnings.)
-			 *
-			 * Autocommit mode is forced by either a true forceCommit
-			 * parameter to me, or a true preventChain parameter to the
-			 * preceding StartTransactionCommand call, or a
-			 * PreventTransactionChain call during the transaction.
-			 * (The parameters could be omitted, but it turns out most
-			 * callers of StartTransactionCommand/CommitTransactionCommand
-			 * want to force autocommit, so making them all call
-			 * PreventTransactionChain would just be extra notation.)
+			 * If we aren't in a transaction block, just do our usual
+			 * transaction commit.
 			 */
 		case TBLOCK_DEFAULT:
-			if (autocommit || forceCommit || suppressChain)
-				CommitTransaction();
-			else
-			{
-				BeginTransactionBlock();
-				Assert(s->blockState == TBLOCK_INPROGRESS);
-				/* This code must match the TBLOCK_INPROGRESS case below: */
-				CommandCounterIncrement();
-				MemoryContextResetAndDeleteChildren(TransactionCommandContext);
-			}
+			CommitTransaction();
 			break;
 
 			/*
@@ -1357,15 +1271,10 @@ CommitTransactionCommand(bool forceCommit)
 			/*
 			 * This is the case when we have finished executing a command
 			 * someplace within a transaction block.  We increment the
-			 * command counter and return.	Someday we may free resources
-			 * local to the command.
-			 *
-			 * That someday is today, at least for memory allocated in
-			 * TransactionCommandContext. - vadim 03/25/97
+			 * command counter and return.
 			 */
 		case TBLOCK_INPROGRESS:
 			CommandCounterIncrement();
-			MemoryContextResetAndDeleteChildren(TransactionCommandContext);
 			break;
 
 			/*
@@ -1399,9 +1308,8 @@ CommitTransactionCommand(bool forceCommit)
 	}
 }
 
-/* --------------------------------
- *		AbortCurrentTransaction
- * --------------------------------
+/*
+ *	AbortCurrentTransaction
  */
 void
 AbortCurrentTransaction(void)
@@ -1476,24 +1384,21 @@ AbortCurrentTransaction(void)
 	}
 }
 
-/* --------------------------------
- *		PreventTransactionChain
+/*
+ *	PreventTransactionChain
  *
- * This routine is to be called by statements that must not run inside
- * a transaction block, typically because they have non-rollback-able
- * side effects or do internal commits.
+ *	This routine is to be called by statements that must not run inside
+ *	a transaction block, typically because they have non-rollback-able
+ *	side effects or do internal commits.
  *
- * If we have already started a transaction block, issue an error; also issue
- * an error if we appear to be running inside a user-defined function (which
- * could issue more commands and possibly cause a failure after the statement
- * completes).  In autocommit-off mode, we allow the statement if a block is
- * not already started, and force the statement to be autocommitted despite
- * the mode.
+ *	If we have already started a transaction block, issue an error; also issue
+ *	an error if we appear to be running inside a user-defined function (which
+ *	could issue more commands and possibly cause a failure after the statement
+ *	completes).
  *
- * stmtNode: pointer to parameter block for statement; this is used in
- * a very klugy way to determine whether we are inside a function.
- * stmtType: statement type name for error messages.
- * --------------------------------
+ *	stmtNode: pointer to parameter block for statement; this is used in
+ *	a very klugy way to determine whether we are inside a function.
+ *	stmtType: statement type name for error messages.
  */
 void
 PreventTransactionChain(void *stmtNode, const char *stmtType)
@@ -1502,34 +1407,29 @@ PreventTransactionChain(void *stmtNode, const char *stmtType)
 	 * xact block already started?
 	 */
 	if (IsTransactionBlock())
-	{
+		ereport(ERROR,
+				(errcode(ERRCODE_ACTIVE_SQL_TRANSACTION),
 		/* translator: %s represents an SQL statement name */
-		elog(ERROR, "%s cannot run inside a transaction block", stmtType);
-	}
+				 errmsg("%s cannot run inside a transaction block",
+						stmtType)));
+
 	/*
 	 * Are we inside a function call?  If the statement's parameter block
 	 * was allocated in QueryContext, assume it is an interactive command.
 	 * Otherwise assume it is coming from a function.
 	 */
 	if (!MemoryContextContains(QueryContext, stmtNode))
-	{
+		ereport(ERROR,
+				(errcode(ERRCODE_ACTIVE_SQL_TRANSACTION),
 		/* translator: %s represents an SQL statement name */
-		elog(ERROR, "%s cannot be executed from a function", stmtType);
-	}
+			 errmsg("%s cannot be executed from a function", stmtType)));
 	/* If we got past IsTransactionBlock test, should be in default state */
 	if (CurrentTransactionState->blockState != TBLOCK_DEFAULT)
-		elog(ERROR, "PreventTransactionChain: can't prevent chain");
-	/* okay to set the flag */
-	suppressChain = true;
-	/* If we're in autocommit-off node, generate a notice */
-	if (!autocommit)
-	{
-		/* translator: %s represents an SQL statement name */
-		elog(NOTICE, "%s will be committed automatically", stmtType);
-	}
+		elog(ERROR, "cannot prevent transaction chain");
+	/* all okay */
 }
 
-/* --------------------------------
+/*
  *	RequireTransactionChain
  *
  *	This routine is to be called by statements that must run inside
@@ -1545,7 +1445,6 @@ PreventTransactionChain(void *stmtNode, const char *stmtType)
  *	stmtNode: pointer to parameter block for statement; this is used in
  *	a very klugy way to determine whether we are inside a function.
  *	stmtType: statement type name for error messages.
- * --------------------------------
  */
 void
 RequireTransactionChain(void *stmtNode, const char *stmtType)
@@ -1555,6 +1454,7 @@ RequireTransactionChain(void *stmtNode, const char *stmtType)
 	 */
 	if (IsTransactionBlock())
 		return;
+
 	/*
 	 * Are we inside a function call?  If the statement's parameter block
 	 * was allocated in QueryContext, assume it is an interactive command.
@@ -1562,15 +1462,67 @@ RequireTransactionChain(void *stmtNode, const char *stmtType)
 	 */
 	if (!MemoryContextContains(QueryContext, stmtNode))
 		return;
-	/*
-	 * If we are in autocommit-off mode then it's okay, because this
-	 * statement will itself start a transaction block.
-	 */
-	if (!autocommit && !suppressChain)
-		return;
+	ereport(ERROR,
+			(errcode(ERRCODE_NO_ACTIVE_SQL_TRANSACTION),
 	/* translator: %s represents an SQL statement name */
-	elog(ERROR, "%s may only be used in begin/end transaction blocks",
-		 stmtType);
+			 errmsg("%s may only be used in transaction blocks",
+					stmtType)));
+}
+
+
+/*
+ * Register or deregister callback functions for end-of-xact cleanup
+ *
+ * These functions are intended for use by dynamically loaded modules.
+ * For built-in modules we generally just hardwire the appropriate calls
+ * (mainly because it's easier to control the order that way, where needed).
+ *
+ * Note that the callback occurs post-commit or post-abort, so the callback
+ * functions can only do noncritical cleanup.
+ */
+void
+RegisterEOXactCallback(EOXactCallback callback, void *arg)
+{
+	EOXactCallbackItem *item;
+
+	item = (EOXactCallbackItem *)
+		MemoryContextAlloc(TopMemoryContext, sizeof(EOXactCallbackItem));
+	item->callback = callback;
+	item->arg = arg;
+	item->next = EOXact_callbacks;
+	EOXact_callbacks = item;
+}
+
+void
+UnregisterEOXactCallback(EOXactCallback callback, void *arg)
+{
+	EOXactCallbackItem *item;
+	EOXactCallbackItem *prev;
+
+	prev = NULL;
+	for (item = EOXact_callbacks; item; prev = item, item = item->next)
+	{
+		if (item->callback == callback && item->arg == arg)
+		{
+			if (prev)
+				prev->next = item->next;
+			else
+				EOXact_callbacks = item->next;
+			pfree(item);
+			break;
+		}
+	}
+}
+
+static void
+CallEOXactCallbacks(bool isCommit)
+{
+	EOXactCallbackItem *item;
+
+	for (item = EOXact_callbacks; item; item = item->next)
+	{
+		(*item->callback) (isCommit, item->arg);
+	}
 }
 
 
@@ -1578,9 +1530,8 @@ RequireTransactionChain(void *stmtNode, const char *stmtType)
  *					   transaction block support
  * ----------------------------------------------------------------
  */
-/* --------------------------------
- *		BeginTransactionBlock
- * --------------------------------
+/*
+ *	BeginTransactionBlock
  */
 void
 BeginTransactionBlock(void)
@@ -1591,7 +1542,9 @@ BeginTransactionBlock(void)
 	 * check the current transaction state
 	 */
 	if (s->blockState != TBLOCK_DEFAULT)
-		elog(WARNING, "BEGIN: already a transaction in progress");
+		ereport(WARNING,
+				(errcode(ERRCODE_ACTIVE_SQL_TRANSACTION),
+				 errmsg("there is already a transaction in progress")));
 
 	/*
 	 * set the current transaction block state information appropriately
@@ -1600,10 +1553,7 @@ BeginTransactionBlock(void)
 	s->blockState = TBLOCK_BEGIN;
 
 	/*
-	 * do begin processing.  NOTE: if you put anything here, check that it
-	 * behaves properly in both autocommit-on and autocommit-off modes. In
-	 * the latter case we will already have done some work in the new
-	 * transaction.
+	 * do begin processing here.  Nothing to do at present.
 	 */
 
 	/*
@@ -1612,9 +1562,8 @@ BeginTransactionBlock(void)
 	s->blockState = TBLOCK_INPROGRESS;
 }
 
-/* --------------------------------
- *		EndTransactionBlock
- * --------------------------------
+/*
+ *	EndTransactionBlock
  */
 void
 EndTransactionBlock(void)
@@ -1655,14 +1604,15 @@ EndTransactionBlock(void)
 	 * CommitTransactionCommand() will then put us back into the default
 	 * state.
 	 */
-	elog(WARNING, "COMMIT: no transaction in progress");
+	ereport(WARNING,
+			(errcode(ERRCODE_NO_ACTIVE_SQL_TRANSACTION),
+			 errmsg("there is no transaction in progress")));
 	AbortTransaction();
 	s->blockState = TBLOCK_ENDABORT;
 }
 
-/* --------------------------------
- *		AbortTransactionBlock
- * --------------------------------
+/*
+ *	AbortTransactionBlock
  */
 #ifdef NOT_USED
 static void
@@ -1692,15 +1642,16 @@ AbortTransactionBlock(void)
 	 * CommitTransactionCommand() will then put us back into the default
 	 * state.
 	 */
-	elog(WARNING, "ROLLBACK: no transaction in progress");
+	ereport(WARNING,
+			(errcode(ERRCODE_NO_ACTIVE_SQL_TRANSACTION),
+			 errmsg("there is no transaction in progress")));
 	AbortTransaction();
 	s->blockState = TBLOCK_ENDABORT;
 }
 #endif
 
-/* --------------------------------
- *		UserAbortTransactionBlock
- * --------------------------------
+/*
+ *	UserAbortTransactionBlock
  */
 void
 UserAbortTransactionBlock(void)
@@ -1739,18 +1690,19 @@ UserAbortTransactionBlock(void)
 	 * CommitTransactionCommand() will then put us back into the default
 	 * state.
 	 */
-	elog(WARNING, "ROLLBACK: no transaction in progress");
+	ereport(WARNING,
+			(errcode(ERRCODE_NO_ACTIVE_SQL_TRANSACTION),
+			 errmsg("there is no transaction in progress")));
 	AbortTransaction();
 	s->blockState = TBLOCK_ENDABORT;
 }
 
-/* --------------------------------
- *		AbortOutOfAnyTransaction
+/*
+ *	AbortOutOfAnyTransaction
  *
- * This routine is provided for error recovery purposes.  It aborts any
- * active transaction or transaction block, leaving the system in a known
- * idle state.
- * --------------------------------
+ *	This routine is provided for error recovery purposes.  It aborts any
+ *	active transaction or transaction block, leaving the system in a known
+ *	idle state.
  */
 void
 AbortOutOfAnyTransaction(void)
@@ -1784,19 +1736,70 @@ AbortOutOfAnyTransaction(void)
 	s->blockState = TBLOCK_DEFAULT;
 }
 
+/*
+ * IsTransactionBlock --- are we within a transaction block?
+ */
 bool
 IsTransactionBlock(void)
 {
 	TransactionState s = CurrentTransactionState;
 
-	if (s->blockState == TBLOCK_INPROGRESS
-		|| s->blockState == TBLOCK_ABORT
-		|| s->blockState == TBLOCK_ENDABORT)
-		return true;
+	if (s->blockState == TBLOCK_DEFAULT)
+		return false;
 
-	return false;
+	return true;
 }
 
+/*
+ * IsTransactionOrTransactionBlock --- are we within either a transaction
+ * or a transaction block?  (The backend is only really "idle" when this
+ * returns false.)
+ *
+ * This should match up with IsTransactionBlock and IsTransactionState.
+ */
+bool
+IsTransactionOrTransactionBlock(void)
+{
+	TransactionState s = CurrentTransactionState;
+
+	if (s->blockState == TBLOCK_DEFAULT && s->state == TRANS_DEFAULT)
+		return false;
+
+	return true;
+}
+
+/*
+ * TransactionBlockStatusCode - return status code to send in ReadyForQuery
+ */
+char
+TransactionBlockStatusCode(void)
+{
+	TransactionState s = CurrentTransactionState;
+
+	switch (s->blockState)
+	{
+		case TBLOCK_DEFAULT:
+			return 'I';			/* idle --- not in transaction */
+		case TBLOCK_BEGIN:
+		case TBLOCK_INPROGRESS:
+		case TBLOCK_END:
+			return 'T';			/* in transaction */
+		case TBLOCK_ABORT:
+		case TBLOCK_ENDABORT:
+			return 'E';			/* in failed transaction */
+	}
+
+	/* should never get here */
+	elog(ERROR, "invalid transaction block state: %d",
+		 (int) s->blockState);
+	return 0;					/* keep compiler quiet */
+}
+
+
+/*
+ *	XLOG support routines
+ */
+
 void
 xact_redo(XLogRecPtr lsn, XLogRecord *record)
 {
@@ -1871,6 +1874,7 @@ XactPopRollback(void)
 {
 	_RollbackFunc = NULL;
 }
+
 #ifdef USE_REPLICATION
 void
 PGR_Reload_Start_Time(void)
