@@ -6,12 +6,12 @@
  * There is hardly anything left of Paul Brown's original implementation...
  *
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994-5, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/commands/cluster.c,v 1.91.2.1 2003/03/03 04:37:48 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/commands/cluster.c,v 1.116.2.1 2004/08/31 23:16:36 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -20,19 +20,22 @@
 #include "access/genam.h"
 #include "access/heapam.h"
 #include "catalog/catalog.h"
+#include "catalog/catname.h"
 #include "catalog/dependency.h"
 #include "catalog/heap.h"
 #include "catalog/index.h"
 #include "catalog/indexing.h"
-#include "catalog/catname.h"
+#include "catalog/namespace.h"
 #include "commands/cluster.h"
 #include "commands/tablecmds.h"
 #include "miscadmin.h"
+#include "utils/acl.h"
 #include "utils/fmgroids.h"
 #include "utils/lsyscache.h"
 #include "utils/syscache.h"
 #include "utils/relcache.h"
 
+
 /*
  * We need one of these structs for each index in the relation to be
  * clustered.  It's basically the data needed by index_create() so
@@ -48,15 +51,190 @@ typedef struct
 	bool		isclustered;
 } IndexAttrs;
 
+/*
+ * This struct is used to pass around the information on tables to be
+ * clustered. We need this so we can make a list of them when invoked without
+ * a specific table/index pair.
+ */
+typedef struct
+{
+	Oid			tableOid;
+	Oid			indexOid;
+} RelToCluster;
+
+
+static void cluster_rel(RelToCluster *rv, bool recheck);
 static Oid	make_new_heap(Oid OIDOldHeap, const char *NewName);
 static void copy_heap_data(Oid OIDNewHeap, Oid OIDOldHeap, Oid OIDOldIndex);
 static List *get_indexattr_list(Relation OldHeap, Oid OldIndex);
-static void recreate_indexattr(Oid OIDOldHeap, List *indexes);
+static void rebuild_indexes(Oid OIDOldHeap, List *indexes);
 static void swap_relfilenodes(Oid r1, Oid r2);
+static List *get_tables_to_cluster(MemoryContext cluster_context);
 
 
+
+/*---------------------------------------------------------------------------
+ * This cluster code allows for clustering multiple tables at once. Because
+ * of this, we cannot just run everything on a single transaction, or we
+ * would be forced to acquire exclusive locks on all the tables being
+ * clustered, simultaneously --- very likely leading to deadlock.
+ *
+ * To solve this we follow a similar strategy to VACUUM code,
+ * clustering each relation in a separate transaction. For this to work,
+ * we need to:
+ *	- provide a separate memory context so that we can pass information in
+ *	  a way that survives across transactions
+ *	- start a new transaction every time a new relation is clustered
+ *	- check for validity of the information on to-be-clustered relations,
+ *	  as someone might have deleted a relation behind our back, or
+ *	  clustered one on a different index
+ *	- end the transaction
+ *
+ * The single-relation case does not have any such overhead.
+ *
+ * We also allow a relation being specified without index.	In that case,
+ * the indisclustered bit will be looked up, and an ERROR will be thrown
+ * if there is no index with the bit set.
+ *---------------------------------------------------------------------------
+ */
+void
+cluster(ClusterStmt *stmt)
+{
+	if (stmt->relation != NULL)
+	{
+		/* This is the single-relation case. */
+		Oid			tableOid,
+					indexOid = InvalidOid;
+		Relation	rel;
+		RelToCluster rvtc;
+
+		/* Find and lock the table */
+		rel = heap_openrv(stmt->relation, AccessExclusiveLock);
+
+		tableOid = RelationGetRelid(rel);
+
+		/* Check permissions */
+		if (!pg_class_ownercheck(tableOid, GetUserId()))
+			aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_CLASS,
+						   RelationGetRelationName(rel));
+
+		if (stmt->indexname == NULL)
+		{
+			List	   *index;
+
+			/* We need to find the index that has indisclustered set. */
+			foreach(index, RelationGetIndexList(rel))
+			{
+				HeapTuple	idxtuple;
+				Form_pg_index indexForm;
+
+				indexOid = lfirsto(index);
+				idxtuple = SearchSysCache(INDEXRELID,
+										  ObjectIdGetDatum(indexOid),
+										  0, 0, 0);
+				if (!HeapTupleIsValid(idxtuple))
+					elog(ERROR, "cache lookup failed for index %u", indexOid);
+				indexForm = (Form_pg_index) GETSTRUCT(idxtuple);
+				if (indexForm->indisclustered)
+				{
+					ReleaseSysCache(idxtuple);
+					break;
+				}
+				ReleaseSysCache(idxtuple);
+				indexOid = InvalidOid;
+			}
+
+			if (!OidIsValid(indexOid))
+				ereport(ERROR,
+						(errcode(ERRCODE_UNDEFINED_OBJECT),
+						 errmsg("there is no previously clustered index for table \"%s\"",
+								stmt->relation->relname)));
+		}
+		else
+		{
+			/*
+			 * The index is expected to be in the same namespace as the
+			 * relation.
+			 */
+			indexOid = get_relname_relid(stmt->indexname,
+										 rel->rd_rel->relnamespace);
+			if (!OidIsValid(indexOid))
+				ereport(ERROR,
+						(errcode(ERRCODE_UNDEFINED_OBJECT),
+				   errmsg("index \"%s\" for table \"%s\" does not exist",
+						  stmt->indexname, stmt->relation->relname)));
+		}
+
+		/* All other checks are done in cluster_rel() */
+		rvtc.tableOid = tableOid;
+		rvtc.indexOid = indexOid;
+
+		/* close relation, keep lock till commit */
+		heap_close(rel, NoLock);
+
+		/* Do the job */
+		cluster_rel(&rvtc, false);
+	}
+	else
+	{
+		/*
+		 * This is the "multi relation" case. We need to cluster all
+		 * tables that have some index with indisclustered set.
+		 */
+		MemoryContext cluster_context;
+		List	   *rv,
+				   *rvs;
+
+		/*
+		 * We cannot run this form of CLUSTER inside a user transaction
+		 * block; we'd be holding locks way too long.
+		 */
+		PreventTransactionChain((void *) stmt, "CLUSTER");
+
+		/*
+		 * Create special memory context for cross-transaction storage.
+		 *
+		 * Since it is a child of PortalContext, it will go away even in case
+		 * of error.
+		 */
+		cluster_context = AllocSetContextCreate(PortalContext,
+												"Cluster",
+												ALLOCSET_DEFAULT_MINSIZE,
+												ALLOCSET_DEFAULT_INITSIZE,
+												ALLOCSET_DEFAULT_MAXSIZE);
+
+		/*
+		 * Build the list of relations to cluster.	Note that this lives
+		 * in cluster_context.
+		 */
+		rvs = get_tables_to_cluster(cluster_context);
+
+		/* Commit to get out of starting transaction */
+		CommitTransactionCommand();
+
+		/* Ok, now that we've got them all, cluster them one by one */
+		foreach(rv, rvs)
+		{
+			RelToCluster *rvtc = (RelToCluster *) lfirst(rv);
+
+			/* Start a new transaction for each relation. */
+			StartTransactionCommand();
+			SetQuerySnapshot(); /* might be needed for functions in
+								 * indexes */
+			cluster_rel(rvtc, true);
+			CommitTransactionCommand();
+		}
+
+		/* Start a new transaction for the cleanup work. */
+		StartTransactionCommand();
+
+		/* Clean up working storage */
+		MemoryContextDelete(cluster_context);
+	}
+}
+
 /*
- * cluster
+ * cluster_rel
  *
  * This clusters the table by creating a new, clustered table and
  * swapping the relfilenodes of the new table and the old table, so
@@ -68,74 +246,124 @@ static void swap_relfilenodes(Oid r1, Oid r2);
  * same way we do for the relation.  Since we are effectively bulk-loading
  * the new table, it's better to create the indexes afterwards than to fill
  * them incrementally while we load the table.
- *
- * Permissions checks were done already.
  */
-void
-cluster(RangeVar *oldrelation, char *oldindexname)
+static void
+cluster_rel(RelToCluster *rvtc, bool recheck)
 {
-	Oid			OIDOldHeap,
-				OIDOldIndex,
-				OIDNewHeap;
 	Relation	OldHeap,
 				OldIndex;
-	char		NewHeapName[NAMEDATALEN];
-	ObjectAddress object;
-	List	   *indexes;
+
+	/* Check for user-requested abort. */
+	CHECK_FOR_INTERRUPTS();
+
+	/*
+	 * Since we may open a new transaction for each relation, we have to
+	 * check that the relation still is what we think it is.
+	 *
+	 * If this is a single-transaction CLUSTER, we can skip these tests. We
+	 * *must* skip the one on indisclustered since it would reject an
+	 * attempt to cluster a not-previously-clustered index.
+	 */
+	if (recheck)
+	{
+		HeapTuple	tuple;
+		Form_pg_index indexForm;
+
+		/*
+		 * Check if the relation and index still exist before opening them
+		 */
+		if (!SearchSysCacheExists(RELOID,
+								  ObjectIdGetDatum(rvtc->tableOid),
+								  0, 0, 0) ||
+			!SearchSysCacheExists(RELOID,
+								  ObjectIdGetDatum(rvtc->indexOid),
+								  0, 0, 0))
+			return;
+
+		/* Check that the user still owns the relation */
+		if (!pg_class_ownercheck(rvtc->tableOid, GetUserId()))
+			return;
+
+		/*
+		 * Check that the index is still the one with indisclustered set.
+		 */
+		tuple = SearchSysCache(INDEXRELID,
+							   ObjectIdGetDatum(rvtc->indexOid),
+							   0, 0, 0);
+		if (!HeapTupleIsValid(tuple))
+			return;				/* could have gone away... */
+		indexForm = (Form_pg_index) GETSTRUCT(tuple);
+		if (!indexForm->indisclustered)
+		{
+			ReleaseSysCache(tuple);
+			return;
+		}
+		ReleaseSysCache(tuple);
+	}
 
 	/*
 	 * We grab exclusive access to the target rel and index for the
-	 * duration of the transaction.
+	 * duration of the transaction.  (This is redundant for the single-
+	 * transaction case, since cluster() already did it.)
 	 */
-	OldHeap = heap_openrv(oldrelation, AccessExclusiveLock);
-	OIDOldHeap = RelationGetRelid(OldHeap);
+	OldHeap = heap_open(rvtc->tableOid, AccessExclusiveLock);
 
-	/*
-	 * The index is expected to be in the same namespace as the relation.
-	 */
-	OIDOldIndex = get_relname_relid(oldindexname,
-									RelationGetNamespace(OldHeap));
-	if (!OidIsValid(OIDOldIndex))
-		elog(ERROR, "CLUSTER: cannot find index \"%s\" for table \"%s\"",
-			 oldindexname, RelationGetRelationName(OldHeap));
-	OldIndex = index_open(OIDOldIndex);
+	OldIndex = index_open(rvtc->indexOid);
 	LockRelation(OldIndex, AccessExclusiveLock);
 
 	/*
 	 * Check that index is in fact an index on the given relation
 	 */
 	if (OldIndex->rd_index == NULL ||
-		OldIndex->rd_index->indrelid != OIDOldHeap)
-		elog(ERROR, "CLUSTER: \"%s\" is not an index for table \"%s\"",
-			 RelationGetRelationName(OldIndex),
-			 RelationGetRelationName(OldHeap));
+		OldIndex->rd_index->indrelid != rvtc->tableOid)
+		ereport(ERROR,
+				(errcode(ERRCODE_WRONG_OBJECT_TYPE),
+				 errmsg("\"%s\" is not an index for table \"%s\"",
+						RelationGetRelationName(OldIndex),
+						RelationGetRelationName(OldHeap))));
 
 	/*
-	 * Disallow clustering on incomplete indexes (those that might not index
-	 * every row of the relation).  We could relax this by making a separate
-	 * seqscan pass over the table to copy the missing rows, but that seems
-	 * expensive and tedious.
+	 * Disallow clustering on incomplete indexes (those that might not
+	 * index every row of the relation).  We could relax this by making a
+	 * separate seqscan pass over the table to copy the missing rows, but
+	 * that seems expensive and tedious.
 	 */
-	if (VARSIZE(&OldIndex->rd_index->indpred) > VARHDRSZ) /* partial? */
-		elog(ERROR, "CLUSTER: cannot cluster on partial index");
+	if (!heap_attisnull(OldIndex->rd_indextuple, Anum_pg_index_indpred))
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("cannot cluster on partial index")));
 	if (!OldIndex->rd_am->amindexnulls)
 	{
 		AttrNumber	colno;
 
 		/*
 		 * If the AM doesn't index nulls, then it's a partial index unless
-		 * we can prove all the rows are non-null.  Note we only need look
+		 * we can prove all the rows are non-null.	Note we only need look
 		 * at the first column; multicolumn-capable AMs are *required* to
 		 * index nulls in columns after the first.
 		 */
-		if (OidIsValid(OldIndex->rd_index->indproc))
-			elog(ERROR, "CLUSTER: cannot cluster on functional index when index access method does not handle nulls");
 		colno = OldIndex->rd_index->indkey[0];
-		if (colno > 0)			/* system columns are non-null */
+		if (colno > 0)
+		{
+			/* ordinary user attribute */
 			if (!OldHeap->rd_att->attrs[colno - 1]->attnotnull)
-				elog(ERROR, "CLUSTER: cannot cluster when index access method does not handle nulls"
-					 "\n\tYou may be able to work around this by marking column \"%s\" NOT NULL",
-					 NameStr(OldHeap->rd_att->attrs[colno - 1]->attname));
+				ereport(ERROR,
+						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						 errmsg("cannot cluster when index access method does not handle null values"),
+						 errhint("You may be able to work around this by marking column \"%s\" NOT NULL.",
+				  NameStr(OldHeap->rd_att->attrs[colno - 1]->attname))));
+		}
+		else if (colno < 0)
+		{
+			/* system column --- okay, always non-null */
+		}
+		else
+		{
+			/* index expression, lose... */
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+					 errmsg("cannot cluster on expressional index when index access method does not handle null values")));
+		}
 	}
 
 	/*
@@ -146,14 +374,55 @@ cluster(RangeVar *oldrelation, char *oldindexname)
 	 * might work for other system relations, but I ain't gonna risk it.
 	 */
 	if (IsSystemRelation(OldHeap))
-		elog(ERROR, "CLUSTER: cannot cluster system relation \"%s\"",
-			 RelationGetRelationName(OldHeap));
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("\"%s\" is a system catalog",
+						RelationGetRelationName(OldHeap))));
 
-	/* Save the information of all indexes on the relation. */
-	indexes = get_indexattr_list(OldHeap, OIDOldIndex);
+	/*
+	 * Don't allow cluster on temp tables of other backends ... their
+	 * local buffer manager is not going to cope.
+	 */
+	if (isOtherTempNamespace(RelationGetNamespace(OldHeap)))
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			   errmsg("cannot cluster temporary tables of other sessions")));
 
-	/* Drop relcache refcnts, but do NOT give up the locks */
+	/* Drop relcache refcnt on OldIndex, but keep lock */
 	index_close(OldIndex);
+
+	/* rebuild_relation does all the dirty work */
+	rebuild_relation(OldHeap, rvtc->indexOid);
+
+	/* NB: rebuild_relation does heap_close() on OldHeap */
+}
+
+/*
+ * rebuild_relation: rebuild an existing relation
+ *
+ * This is shared code between CLUSTER and TRUNCATE.  In the TRUNCATE
+ * case, the new relation is built and left empty.	In the CLUSTER case,
+ * it is filled with data read from the old relation in the order specified
+ * by the index.
+ *
+ * OldHeap: table to rebuild --- must be opened and exclusive-locked!
+ * indexOid: index to cluster by, or InvalidOid in TRUNCATE case
+ *
+ * NB: this routine closes OldHeap at the right time; caller should not.
+ */
+void
+rebuild_relation(Relation OldHeap, Oid indexOid)
+{
+	Oid			tableOid = RelationGetRelid(OldHeap);
+	List	   *indexes;
+	Oid			OIDNewHeap;
+	char		NewHeapName[NAMEDATALEN];
+	ObjectAddress object;
+
+	/* Save the information about all indexes on the relation. */
+	indexes = get_indexattr_list(OldHeap, indexOid);
+
+	/* Close relcache entry, but keep lock until transaction commit */
 	heap_close(OldHeap, NoLock);
 
 	/*
@@ -164,9 +433,9 @@ cluster(RangeVar *oldrelation, char *oldindexname)
 	 * namespace from the old, or we will have problems with the TEMP
 	 * status of temp tables.
 	 */
-	snprintf(NewHeapName, NAMEDATALEN, "pg_temp_%u", OIDOldHeap);
+	snprintf(NewHeapName, sizeof(NewHeapName), "pg_temp_%u", tableOid);
 
-	OIDNewHeap = make_new_heap(OIDOldHeap, NewHeapName);
+	OIDNewHeap = make_new_heap(tableOid, NewHeapName);
 
 	/*
 	 * We don't need CommandCounterIncrement() because make_new_heap did
@@ -176,13 +445,14 @@ cluster(RangeVar *oldrelation, char *oldindexname)
 	/*
 	 * Copy the heap data into the new table in the desired order.
 	 */
-	copy_heap_data(OIDNewHeap, OIDOldHeap, OIDOldIndex);
+	if (OidIsValid(indexOid))
+		copy_heap_data(OIDNewHeap, tableOid, indexOid);
 
 	/* To make the new heap's data visible (probably not needed?). */
 	CommandCounterIncrement();
 
 	/* Swap the relfilenodes of the old and new heaps. */
-	swap_relfilenodes(OIDOldHeap, OIDNewHeap);
+	swap_relfilenodes(tableOid, OIDNewHeap);
 
 	CommandCounterIncrement();
 
@@ -201,9 +471,9 @@ cluster(RangeVar *oldrelation, char *oldindexname)
 
 	/*
 	 * Recreate each index on the relation.  We do not need
-	 * CommandCounterIncrement() because recreate_indexattr does it.
+	 * CommandCounterIncrement() because rebuild_indexes does it.
 	 */
-	recreate_indexattr(OIDOldHeap, indexes);
+	rebuild_indexes(tableOid, indexes);
 }
 
 /*
@@ -231,6 +501,7 @@ make_new_heap(Oid OIDOldHeap, const char *NewName)
 										  tupdesc,
 										  OldHeap->rd_rel->relkind,
 										  OldHeap->rd_rel->relisshared,
+										  ONCOMMIT_NOOP,
 										  allowSystemTableMods);
 
 	/*
@@ -314,43 +585,25 @@ get_indexattr_list(Relation OldHeap, Oid OldIndex)
 	/* Ask the relcache to produce a list of the indexes of the old rel */
 	foreach(indlist, RelationGetIndexList(OldHeap))
 	{
-		Oid			indexOID = (Oid) lfirsti(indlist);
-		HeapTuple	indexTuple;
-		HeapTuple	classTuple;
-		Form_pg_index indexForm;
-		Form_pg_class classForm;
+		Oid			indexOID = lfirsto(indlist);
+		Relation	oldIndex;
 		IndexAttrs *attrs;
 
-		indexTuple = SearchSysCache(INDEXRELID,
-									ObjectIdGetDatum(indexOID),
-									0, 0, 0);
-		if (!HeapTupleIsValid(indexTuple))
-			elog(ERROR, "Cache lookup failed for index %u", indexOID);
-		indexForm = (Form_pg_index) GETSTRUCT(indexTuple);
-		Assert(indexForm->indexrelid == indexOID);
+		oldIndex = index_open(indexOID);
 
 		attrs = (IndexAttrs *) palloc(sizeof(IndexAttrs));
 		attrs->indexOID = indexOID;
-		attrs->indexInfo = BuildIndexInfo(indexForm);
+		attrs->indexName = pstrdup(NameStr(oldIndex->rd_rel->relname));
+		attrs->accessMethodOID = oldIndex->rd_rel->relam;
+		attrs->indexInfo = BuildIndexInfo(oldIndex);
 		attrs->classOID = (Oid *)
 			palloc(sizeof(Oid) * attrs->indexInfo->ii_NumIndexAttrs);
-		memcpy(attrs->classOID, indexForm->indclass,
+		memcpy(attrs->classOID, oldIndex->rd_index->indclass,
 			   sizeof(Oid) * attrs->indexInfo->ii_NumIndexAttrs);
-		attrs->isclustered = (OldIndex == indexOID);
+		/* We adjust the isclustered attribute to correct new state */
+		attrs->isclustered = (indexOID == OldIndex);
 
-		/* Name and access method of each index come from pg_class */
-		classTuple = SearchSysCache(RELOID,
-									ObjectIdGetDatum(indexOID),
-									0, 0, 0);
-		if (!HeapTupleIsValid(classTuple))
-			elog(ERROR, "Cache lookup failed for index %u", indexOID);
-		classForm = (Form_pg_class) GETSTRUCT(classTuple);
-
-		attrs->indexName = pstrdup(NameStr(classForm->relname));
-		attrs->accessMethodOID = classForm->relam;
-
-		ReleaseSysCache(classTuple);
-		ReleaseSysCache(indexTuple);
+		index_close(oldIndex);
 
 		/*
 		 * Cons the gathered data into the list.  We do not care about
@@ -367,7 +620,7 @@ get_indexattr_list(Relation OldHeap, Oid OldIndex)
  * the new index (carrying the old index filenode along).
  */
 static void
-recreate_indexattr(Oid OIDOldHeap, List *indexes)
+rebuild_indexes(Oid OIDOldHeap, List *indexes)
 {
 	List	   *elem;
 
@@ -382,7 +635,8 @@ recreate_indexattr(Oid OIDOldHeap, List *indexes)
 		Relation	pg_index;
 
 		/* Create the new index under a temporary name */
-		snprintf(newIndexName, NAMEDATALEN, "pg_temp_%u", attrs->indexOID);
+		snprintf(newIndexName, sizeof(newIndexName),
+				 "pg_temp_%u", attrs->indexOID);
 
 		/*
 		 * The new index will have primary and constraint status set to
@@ -462,14 +716,14 @@ swap_relfilenodes(Oid r1, Oid r2)
 								 ObjectIdGetDatum(r1),
 								 0, 0, 0);
 	if (!HeapTupleIsValid(reltup1))
-		elog(ERROR, "CLUSTER: Cannot find tuple for relation %u", r1);
+		elog(ERROR, "cache lookup failed for relation %u", r1);
 	relform1 = (Form_pg_class) GETSTRUCT(reltup1);
 
 	reltup2 = SearchSysCacheCopy(RELOID,
 								 ObjectIdGetDatum(r2),
 								 0, 0, 0);
 	if (!HeapTupleIsValid(reltup2))
-		elog(ERROR, "CLUSTER: Cannot find tuple for relation %u", r2);
+		elog(ERROR, "cache lookup failed for relation %u", r2);
 	relform2 = (Form_pg_class) GETSTRUCT(reltup2);
 
 	/*
@@ -481,13 +735,13 @@ swap_relfilenodes(Oid r1, Oid r2)
 	rel = relation_open(r1, NoLock);
 	i = FlushRelationBuffers(rel, 0);
 	if (i < 0)
-		elog(ERROR, "CLUSTER: FlushRelationBuffers returned %d", i);
+		elog(ERROR, "FlushRelationBuffers returned %d", i);
 	relation_close(rel, NoLock);
 
 	rel = relation_open(r2, NoLock);
 	i = FlushRelationBuffers(rel, 0);
 	if (i < 0)
-		elog(ERROR, "CLUSTER: FlushRelationBuffers returned %d", i);
+		elog(ERROR, "FlushRelationBuffers returned %d", i);
 	relation_close(rel, NoLock);
 
 	/*
@@ -505,8 +759,8 @@ swap_relfilenodes(Oid r1, Oid r2)
 
 	/* swap size statistics too, since new rel has freshly-updated stats */
 	{
-		int4	swap_pages;
-		float4	swap_tuples;
+		int4		swap_pages;
+		float4		swap_tuples;
 
 		swap_pages = relform1->relpages;
 		relform1->relpages = relform2->relpages;
@@ -533,9 +787,8 @@ swap_relfilenodes(Oid r1, Oid r2)
 	 * their new owning relations.	Otherwise the wrong one will get
 	 * dropped ...
 	 *
-	 * NOTE: for now, we can assume the new table will have a TOAST table if
-	 * and only if the old one does.  This logic might need work if we get
-	 * smarter about dropped columns.
+	 * NOTE: it is possible that only one table has a toast table; this can
+	 * happen in CLUSTER if there were dropped columns in the old table.
 	 *
 	 * NOTE: at present, a TOAST table's only dependency is the one on its
 	 * owning table.  If more are ever created, we'd need to use something
@@ -548,35 +801,43 @@ swap_relfilenodes(Oid r1, Oid r2)
 					toastobject;
 		long		count;
 
-		if (!(relform1->reltoastrelid && relform2->reltoastrelid))
-			elog(ERROR, "CLUSTER: expected both swapped tables to have TOAST tables");
-
 		/* Delete old dependencies */
-		count = deleteDependencyRecordsFor(RelOid_pg_class,
-										   relform1->reltoastrelid);
-		if (count != 1)
-			elog(ERROR, "CLUSTER: expected one dependency record for TOAST table, found %ld",
-				 count);
-		count = deleteDependencyRecordsFor(RelOid_pg_class,
-										   relform2->reltoastrelid);
-		if (count != 1)
-			elog(ERROR, "CLUSTER: expected one dependency record for TOAST table, found %ld",
-				 count);
+		if (relform1->reltoastrelid)
+		{
+			count = deleteDependencyRecordsFor(RelOid_pg_class,
+											   relform1->reltoastrelid);
+			if (count != 1)
+				elog(ERROR, "expected one dependency record for TOAST table, found %ld",
+					 count);
+		}
+		if (relform2->reltoastrelid)
+		{
+			count = deleteDependencyRecordsFor(RelOid_pg_class,
+											   relform2->reltoastrelid);
+			if (count != 1)
+				elog(ERROR, "expected one dependency record for TOAST table, found %ld",
+					 count);
+		}
 
 		/* Register new dependencies */
 		baseobject.classId = RelOid_pg_class;
-		baseobject.objectId = r1;
 		baseobject.objectSubId = 0;
 		toastobject.classId = RelOid_pg_class;
-		toastobject.objectId = relform1->reltoastrelid;
 		toastobject.objectSubId = 0;
 
-		recordDependencyOn(&toastobject, &baseobject, DEPENDENCY_INTERNAL);
+		if (relform1->reltoastrelid)
+		{
+			baseobject.objectId = r1;
+			toastobject.objectId = relform1->reltoastrelid;
+			recordDependencyOn(&toastobject, &baseobject, DEPENDENCY_INTERNAL);
+		}
 
-		baseobject.objectId = r2;
-		toastobject.objectId = relform2->reltoastrelid;
-
-		recordDependencyOn(&toastobject, &baseobject, DEPENDENCY_INTERNAL);
+		if (relform2->reltoastrelid)
+		{
+			baseobject.objectId = r2;
+			toastobject.objectId = relform2->reltoastrelid;
+			recordDependencyOn(&toastobject, &baseobject, DEPENDENCY_INTERNAL);
+		}
 	}
 
 	/*
@@ -598,3 +859,60 @@ swap_relfilenodes(Oid r1, Oid r2)
 
 	heap_close(relRelation, RowExclusiveLock);
 }
+
+/*
+ * Get a list of tables that the current user owns and
+ * have indisclustered set.  Return the list in a List * of rvsToCluster
+ * with the tableOid and the indexOid on which the table is already
+ * clustered.
+ */
+static List *
+get_tables_to_cluster(MemoryContext cluster_context)
+{
+	Relation	indRelation;
+	HeapScanDesc scan;
+	ScanKeyData entry;
+	HeapTuple	indexTuple;
+	Form_pg_index index;
+	MemoryContext old_context;
+	RelToCluster *rvtc;
+	List	   *rvs = NIL;
+
+	/*
+	 * Get all indexes that have indisclustered set and are owned by
+	 * appropriate user. System relations or nailed-in relations cannot
+	 * ever have indisclustered set, because CLUSTER will refuse to set it
+	 * when called with one of them as argument.
+	 */
+	indRelation = relation_openr(IndexRelationName, AccessShareLock);
+	ScanKeyEntryInitialize(&entry, 0,
+						   Anum_pg_index_indisclustered,
+						   F_BOOLEQ,
+						   BoolGetDatum(true));
+	scan = heap_beginscan(indRelation, SnapshotNow, 1, &entry);
+	while ((indexTuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
+	{
+		index = (Form_pg_index) GETSTRUCT(indexTuple);
+
+		if (!pg_class_ownercheck(index->indrelid, GetUserId()))
+			continue;
+
+		/*
+		 * We have to build the list in a different memory context so it
+		 * will survive the cross-transaction processing
+		 */
+		old_context = MemoryContextSwitchTo(cluster_context);
+
+		rvtc = (RelToCluster *) palloc(sizeof(RelToCluster));
+		rvtc->tableOid = index->indrelid;
+		rvtc->indexOid = index->indexrelid;
+		rvs = lcons(rvtc, rvs);
+
+		MemoryContextSwitchTo(old_context);
+	}
+	heap_endscan(scan);
+
+	relation_close(indRelation, AccessShareLock);
+
+	return rvs;
+}
