@@ -3,12 +3,12 @@
  * proc.c
  *	  routines to manage per-process shared memory data structure
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/storage/lmgr/proc.c,v 1.127 2002/10/31 21:34:16 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/storage/lmgr/proc.c,v 1.136 2003/10/16 20:59:35 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -82,6 +82,7 @@ static bool waitingForSignal = false;
 /* Mark these volatile because they can be changed by signal handler */
 static volatile bool statement_timeout_active = false;
 static volatile bool deadlock_timeout_active = false;
+
 /* statement_fin_time is valid only if statement_timeout_active is true */
 static struct timeval statement_fin_time;
 
@@ -152,7 +153,9 @@ InitProcGlobal(int maxBackends)
 
 			proc = (PGPROC *) ShmemAlloc(sizeof(PGPROC));
 			if (!proc)
-				elog(FATAL, "cannot create new proc: out of memory");
+				ereport(FATAL,
+						(errcode(ERRCODE_OUT_OF_MEMORY),
+						 errmsg("out of shared memory")));
 			MemSet(proc, 0, sizeof(PGPROC));
 			PGSemaphoreCreate(&proc->sem);
 			proc->links.next = ProcGlobal->freeProcs;
@@ -166,7 +169,9 @@ InitProcGlobal(int maxBackends)
 		 */
 		DummyProc = (PGPROC *) ShmemAlloc(sizeof(PGPROC));
 		if (!DummyProc)
-			elog(FATAL, "cannot create new proc: out of memory");
+			ereport(FATAL,
+					(errcode(ERRCODE_OUT_OF_MEMORY),
+					 errmsg("out of shared memory")));
 		MemSet(DummyProc, 0, sizeof(PGPROC));
 		DummyProc->pid = 0;		/* marks DummyProc as not in use */
 		PGSemaphoreCreate(&DummyProc->sem);
@@ -193,10 +198,10 @@ InitProcess(void)
 	 * we are a backend, we inherit this by fork() from the postmaster).
 	 */
 	if (procglobal == NULL)
-		elog(PANIC, "InitProcess: Proc Header uninitialized");
+		elog(PANIC, "proc header uninitialized");
 
 	if (MyProc != NULL)
-		elog(ERROR, "InitProcess: you already exist");
+		elog(ERROR, "you already exist");
 
 	/*
 	 * Try to get a proc struct from the free list.  If this fails, we
@@ -220,7 +225,9 @@ InitProcess(void)
 		 * standard error message.
 		 */
 		SpinLockRelease(ProcStructLock);
-		elog(FATAL, "Sorry, too many clients already");
+		ereport(FATAL,
+				(errcode(ERRCODE_TOO_MANY_CONNECTIONS),
+				 errmsg("sorry, too many clients already")));
 	}
 
 	/*
@@ -239,9 +246,6 @@ InitProcess(void)
 	MyProc->lwWaitLink = NULL;
 	MyProc->waitLock = NULL;
 	MyProc->waitHolder = NULL;
-#ifdef USE_REPLICATION
-	MyProc->replicationId = 0;
-#endif
 	SHMQueueInit(&(MyProc->procHolders));
 
 	/*
@@ -277,17 +281,16 @@ InitDummyProcess(void)
 	 * inherit this by fork() from the postmaster).
 	 */
 	if (ProcGlobal == NULL || DummyProc == NULL)
-		elog(PANIC, "InitDummyProcess: Proc Header uninitialized");
+		elog(PANIC, "proc header uninitialized");
 
 	if (MyProc != NULL)
-		elog(ERROR, "InitDummyProcess: you already exist");
+		elog(ERROR, "you already exist");
 
 	/*
 	 * DummyProc should not presently be in use by anyone else
 	 */
 	if (DummyProc->pid != 0)
-		elog(FATAL, "InitDummyProcess: DummyProc is in use by PID %d",
-			 DummyProc->pid);
+		elog(FATAL, "DummyProc is in use by PID %d", DummyProc->pid);
 	MyProc = DummyProc;
 
 	/*
@@ -306,9 +309,6 @@ InitDummyProcess(void)
 	MyProc->lwWaitLink = NULL;
 	MyProc->waitLock = NULL;
 	MyProc->waitHolder = NULL;
-#ifdef USE_REPLICATION
-	MyProc->replicationId=0;
-#endif
 	SHMQueueInit(&(MyProc->procHolders));
 
 	/*
@@ -329,7 +329,7 @@ InitDummyProcess(void)
  * Returns true if we had been waiting for a lock, else false.
  *
  * (Normally, this would only happen if we accept a cancel/die
- * interrupt while waiting; but an elog(ERROR) while waiting is
+ * interrupt while waiting; but an ereport(ERROR) while waiting is
  * within the realm of possibility, too.)
  */
 bool
@@ -531,7 +531,7 @@ int
 ProcSleep(LOCKMETHODTABLE *lockMethodTable,
 		  LOCKMODE lockmode,
 		  LOCK *lock,
-		  PROCLOCK *holder)
+		  PROCLOCK *proclock)
 {
 	LWLockId	masterLock = lockMethodTable->masterLock;
 	PROC_QUEUE *waitQueue = &(lock->waitProcs);
@@ -576,8 +576,10 @@ ProcSleep(LOCKMETHODTABLE *lockMethodTable,
 					 * up correctly is to call RemoveFromWaitQueue(), but
 					 * we can't do that until we are *on* the wait queue.
 					 * So, set a flag to check below, and break out of
-					 * loop.
+					 * loop.  Also, record deadlock info for later
+					 * message.
 					 */
+					RememberSimpleDeadLock(MyProc, lockmode, lock, proc);
 					early_deadlock = true;
 					break;
 				}
@@ -586,26 +588,15 @@ ProcSleep(LOCKMETHODTABLE *lockMethodTable,
 					LockCheckConflicts(lockMethodTable,
 									   lockmode,
 									   lock,
-									   holder,
+									   proclock,
 									   MyProc,
 									   NULL) == STATUS_OK)
 				{
 					/* Skip the wait and just grant myself the lock. */
-					GrantLock(lock, holder, lockmode);
+					GrantLock(lock, proclock, lockmode);
 					return STATUS_OK;
 				}
 				/* Break out of loop to put myself before him */
-#ifdef USE_REPLICATION 
-				if(proc->replicationId == 0 ||
-				   (MyProc->replicationId > proc->replicationId &&
-				    proc->heldLocks & aheadRequests) ) {
-	elog(DEBUG1,"origin's RId = %d , MyProc->RId = %d , skip",proc->replicationId,MyProc->replicationId);
-				  aheadRequests |= (1 << proc->waitLockMode);
-               			  proc = (PGPROC *) MAKE_PTR(proc->links.next);
-				  continue;
-				}
-
-#endif
 				break;
 			}
 			/* Nope, so advance to next waiter */
@@ -621,20 +612,7 @@ ProcSleep(LOCKMETHODTABLE *lockMethodTable,
 	else
 	{
 		/* I hold no locks, so I can't push in front of anyone. */
-#ifdef USE_REPLICATION
-	  proc = (PGPROC *) &(waitQueue->links);
-	  for (i = 0; i < waitQueue->size+1; i++){
-	elog(DEBUG1,"origin's RId = %d , MyProc->RId = %d",proc->replicationId,MyProc->replicationId);
-	    if(proc->replicationId == 0 ||
-	       MyProc->replicationId > proc->replicationId) {
-	      proc= (PGPROC *) MAKE_PTR(proc->links.next);
-	    }else {
-	      break;
-	    }
-	  }
-#else
-		  proc = (PGPROC *) &(waitQueue->links);
-#endif
+		proc = (PGPROC *) &(waitQueue->links);
 	}
 
 	/*
@@ -648,7 +626,7 @@ ProcSleep(LOCKMETHODTABLE *lockMethodTable,
 
 	/* Set up wait information in PGPROC object, too */
 	MyProc->waitLock = lock;
-	MyProc->waitHolder = holder;
+	MyProc->waitHolder = proclock;
 	MyProc->waitLockMode = lockmode;
 
 	MyProc->errType = STATUS_OK;	/* initialize result for success */
@@ -688,7 +666,7 @@ ProcSleep(LOCKMETHODTABLE *lockMethodTable,
 	 * running the rather expensive deadlock-check code in most cases.
 	 */
 	if (!enable_sig_alarm(DeadlockTimeout, false))
-		elog(FATAL, "ProcSleep: Unable to set timer for process wakeup");
+		elog(FATAL, "could not set timer for process wakeup");
 
 	/*
 	 * If someone wakes us between LWLockRelease and PGSemaphoreLock,
@@ -710,7 +688,7 @@ ProcSleep(LOCKMETHODTABLE *lockMethodTable,
 	 * Disable the timer, if it's still running
 	 */
 	if (!disable_sig_alarm(false))
-		elog(FATAL, "ProcSleep: Unable to disable timer for process wakeup");
+		elog(FATAL, "could not disable timer for process wakeup");
 
 	/*
 	 * Now there is nothing for LockWaitCancel to do.
@@ -759,7 +737,7 @@ ProcWakeup(PGPROC *proc, int errType)
 	SHMQueueDelete(&(proc->links));
 	(proc->waitLock->waitProcs.size)--;
 
-	/* Clean up process' state and pass it the o1k/fail signal */
+	/* Clean up process' state and pass it the ok/fail signal */
 	proc->waitLock = NULL;
 	proc->waitHolder = NULL;
 	proc->errType = errType;
@@ -980,9 +958,14 @@ ProcSendSignal(BackendId procId)
 bool
 enable_sig_alarm(int delayms, bool is_statement_timeout)
 {
+#ifdef WIN32
+#warning add Win32 timer
+#else
 	struct timeval fin_time;
+
 #ifndef __BEOS__
 	struct itimerval timeval;
+
 #else
 	bigtime_t	time_interval;
 #endif
@@ -1012,16 +995,16 @@ enable_sig_alarm(int delayms, bool is_statement_timeout)
 		/*
 		 * Begin deadlock timeout with statement-level timeout active
 		 *
-		 * Here, we want to interrupt at the closer of the two timeout
-		 * times.  If fin_time >= statement_fin_time then we need not
-		 * touch the existing timer setting; else set up to interrupt
-		 * at the deadlock timeout time.
+		 * Here, we want to interrupt at the closer of the two timeout times.
+		 * If fin_time >= statement_fin_time then we need not touch the
+		 * existing timer setting; else set up to interrupt at the
+		 * deadlock timeout time.
 		 *
 		 * NOTE: in this case it is possible that this routine will be
 		 * interrupted by the previously-set timer alarm.  This is okay
-		 * because the signal handler will do only what it should do according
-		 * to the state variables.  The deadlock checker may get run earlier
-		 * than normal, but that does no harm.
+		 * because the signal handler will do only what it should do
+		 * according to the state variables.  The deadlock checker may get
+		 * run earlier than normal, but that does no harm.
 		 */
 		deadlock_timeout_active = true;
 		if (fin_time.tv_sec > statement_fin_time.tv_sec ||
@@ -1048,7 +1031,7 @@ enable_sig_alarm(int delayms, bool is_statement_timeout)
 	if (set_alarm(time_interval, B_ONE_SHOT_RELATIVE_ALARM) < 0)
 		return false;
 #endif
-
+#endif
 	return true;
 }
 
@@ -1062,6 +1045,10 @@ enable_sig_alarm(int delayms, bool is_statement_timeout)
 bool
 disable_sig_alarm(bool is_statement_timeout)
 {
+#ifdef WIN32
+#warning add Win32 timer
+#else
+
 	/*
 	 * Always disable the interrupt if it is active; this avoids being
 	 * interrupted by the signal handler and thereby possibly getting
@@ -1101,7 +1088,7 @@ disable_sig_alarm(bool is_statement_timeout)
 		if (!CheckStatementTimeout())
 			return false;
 	}
-
+#endif
 	return true;
 }
 
@@ -1134,6 +1121,9 @@ CheckStatementTimeout(void)
 	else
 	{
 		/* Not time yet, so (re)schedule the interrupt */
+#ifdef WIN32
+#warning add win32 timer
+#else
 #ifndef __BEOS__
 		struct itimerval timeval;
 
@@ -1156,6 +1146,7 @@ CheckStatementTimeout(void)
 			(statement_fin_time.tv_usec - now.tv_usec);
 		if (set_alarm(time_interval, B_ONE_SHOT_RELATIVE_ALARM) < 0)
 			return false;
+#endif
 #endif
 	}
 
