@@ -10,8 +10,6 @@
 #ifndef REPLICATE_COM_H
 #define	REPLICATE_COM_H 1
 
-#include "pg_config.h"
-
 #ifndef _SYS_TYPES_H
 #include <sys/types.h>
 #endif
@@ -34,15 +32,22 @@
 #define DEFAULT_PGRP_LIFECHECK_PORT	(8201)
 #define DEFAULT_PGRP_RLOG_PORT	(8301)
 #define MAX_DB_SERVER	(32)
-
-/*  Replication packet id */
+/**************************
+*                         *
+*   Packet ID definition  *
+*                         *
+***************************/
+/*=========================
+	Replication packet id
+===========================*/
 #define	CMD_SYS_REPLICATE	'R'
-
+/*-------------------------
+	Simple Query
+--------------------------*/
 #define CMD_STS_SET_SESSION_AUTHORIZATION	'S'
 #define	CMD_STS_TRANSACTION	'T'
 #define	CMD_STS_TEMP_TABLE	'E'
 #define	CMD_STS_QUERY	'Q'
-#define	CMD_STS_COPY	'C'
 #define	CMD_STS_OTHER	'O'
 
 #define CMD_TYPE_VACUUM	'V'
@@ -59,18 +64,35 @@
 #define CMD_TYPE_BEGIN	'B'
 #define CMD_TYPE_COMMIT	'E'
 #define CMD_TYPE_ROLLBACK	'R'
+#define CMD_TYPE_SESSION_AUTHORIZATION_BEGIN	'a'
+#define CMD_TYPE_SESSION_AUTHORIZATION_END	'b'
+#define CMD_TYPE_OTHER	'O'
+
+/*----------------------------
+	Copy Command
+------------------------------*/
+#define	CMD_STS_COPY	'C'
+
 #define CMD_TYPE_COPY	'C'
 #define CMD_TYPE_COPY_DATA	'd'
 #define CMD_TYPE_COPY_DATA_END	'e'
-#define CMD_TYPE_CONNECTION_CLOSE	'x'
-#define CMD_TYPE_SESSION_AUTHORIZATION_BEGIN	'a'
-#define CMD_TYPE_SESSION_AUTHORIZATION_END	'b'
-#define CMD_TYPE_CONNECTION_CLOSE	'x'
-#define CMD_TYPE_OTHER	'O'
 
-/* System call packet id */
+/*----------------------------
+	Large Object
+------------------------------*/
+#define	CMD_STS_LARGE_OBJECT	'L'
+
+#define CMD_TYPE_LO_CREATE	'C'
+#define CMD_TYPE_LO_OPEN	'O'
+#define CMD_TYPE_LO_WRITE	'W'
+#define CMD_TYPE_LO_LSEEK	'S'
+#define CMD_TYPE_LO_CLOSE	'X'
+#define CMD_TYPE_LO_UNLINK	'U'
+
+/*=========================
+	System call packet id
+===========================*/
 #define CMD_SYS_CALL		'S'
-#define CMD_SYS_PREREPLICATE		'Z'
 
 #define	CMD_STS_NOTICE	'N'
 #define	CMD_STS_RESPONSE	'R'
@@ -86,13 +108,17 @@
 #define PGR_TRANSACTION_SOCKET	(0)
 #define PGR_QUERY_SOCKET	(1)
 
-
-
 #define	DATA_FREE	(0)
 #define	DATA_INIT	(1)
 #define	DATA_USE	(2)
-#define	DATA_ERR	(90)
-#define	DATA_END	(-1)
+#define	DATA_STOP	(3)
+#define	DATA_TOP	(10)
+#define	DATA_END	(11)
+#define	DATA_TEMP_USE	(13)
+#define	DATA_NEW	(14)
+#define	DATA_ACCEPT	(20)
+#define	DATA_ERR_NOTICE	(98)
+#define	DATA_ERR	(99)
 #define HOSTNAME_MAX_LENGTH     (128)
 #define DBNAME_MAX_LENGTH       (128)
 #define USERNAME_MAX_LENGTH     (128)
@@ -100,7 +126,7 @@
 #define PATH_MAX_LENGTH        (256)
 #define MAX_SERVER_NUM         (128)
 #define MAX_RETRY_TIMES	(3)
-#define MAX_SOCKET_QUEUE	(100000)
+#define MAX_SOCKET_QUEUE	(10000)
 #define TRANSACTION_ERROR_RESULT	"TRANSACTION_ERROR"
 #define REPLICATE_SERVER_SHM_KEY (1020)
 /* target -> replicate */
@@ -123,7 +149,6 @@
 #define RECOVERY_FINISH	(9)
 #define RECOVERY_ERROR_OCCUPIED	(100)
 #define RECOVERY_ERROR_CONNECTION	(101)
-#define RECOVERY_ERROR_TARGET_ONLY	(102)
 #define RECOVERY_ERROR_ANS	(200)
 #define REPLICATION_SERVER_INFO_TAG "Replicate_Server_Info"
 #define HOST_NAME_TAG	"Host_Name"
@@ -132,7 +157,8 @@
 #define LIFECHECK_PORT_TAG  "LifeCheck_Port"
 #define RSYNC_PATH_TAG	"Rsync_Path"
 #define RSYNC_OPTION_TAG	"Rsync_Option"
-#define RSYNC_COMPRESS_TAG "Rsync_Compress"
+#define STATUS_LOG_FILE_TAG	"Status_Log_File"
+#define ERROR_LOG_FILE_TAG	"Error_Log_File"
 #define RECOVERY_INIT	(0)
 #define RECOVERY_PREPARE_START	(1)
 #define RECOVERY_START	(2)
@@ -159,12 +185,11 @@
 #define PGR_SET_CURRENT_TIME_FUNC_NO	(3)
 #define PGR_NOTICE_DEADLOCK_DETECTION_FUNC_NO	(4)
 #define PGR_TRANSACTION_CONFIRM_ANSWER_FUNC_NO	(5)
-
-
-#define PGR_RELIABLE_MODE_DONE_FUNC_NO		(6)
-#define PGR_NOTICE_ABORT_FUNC_NO		(7)
-#define PGR_SET_CURRENT_REPLICATION_QUERY_ID_NO (8)
-#define PGR_QUERY_CONFIRM_ANSWER_FUNC_NO	(9)
+#define PGR_QUERY_CONFIRM_ANSWER_FUNC_NO	(6)
+#define PGR_RELIABLE_MODE_DONE_FUNC_NO		(7)
+#define PGR_NOTICE_ABORT_FUNC_NO		(8)
+#define PGR_GET_OID_FUNC_NO		(9)
+#define PGR_SET_OID_FUNC_NO		(10)
 #define PGR_CMD_ARG_NUM	(10)
 #define PGR_LOCK_CONFLICT_NOTICE_CMD	"PGR_LOCK_CONFLICT_NOTICE_CMD"
 #define PGR_DEADLOCK_DETECT_NOTICE_CMD	"PGR_DEADLOCK_DETECT_NOTICE_CMD"
@@ -179,7 +204,7 @@
 #define COPYBUFSIZ	(4096)
 #define MAX_WORDS	(24)
 #define MAX_WORD_LETTERS	(48)
-#define PGR_MESSAGE_BUFSIZE	(128)
+#define PGR_MESSAGE_BUFSIZE	(256)
 #define INT_LENGTH	(12)
 #define PGR_MAX_COUNTER	(0x0FFFFFFF)
 #define PGR_GET_OVER_FLOW_FILTER	(0xF0000000)
@@ -208,19 +233,77 @@ union semun {
 
 typedef struct ReplicateHeaderType
 {
-	char cmdSys;
+	char cmdSys;	/*
+					R: replication
+					S: system call
+					C: cascade connection
+					L: replication log
+					*/
 	char cmdSts;	/*
-					Q:query 
-					T:transaction
+					-- with replication --
+						S: session in set session authorization
+						E: session in temp table
+						C: session in copy
+						L: session in large object
+						T: session in transaction
+						Q: session in simple query
+						O: session in others
+					-- with system call -- 
+						N: notice 
+						R: response 
+						A: transaction abort
+						P: query suspend
+						D: query done
+					-- with cascade connection -- 
+						U: to upper
+						L: to lower
+					-- with replication log -- 
+						q: delete query log
+						t: delete transaction log
+						r: update query log
+						u: update transaction log
 					*/
 	char cmdType;	/*
-				S:select
-				I:insert
-				D:delete
-				U:update
-				B:begin
-				E:commit/rollback/end
-				O:others
+					-- with replication --
+						-- session in simple query --
+						V: vacuum
+						A: analyze
+						N: reindex
+						S: select
+						X: explain
+						T: set
+						t: reset
+						I: insert
+						D: delete
+						U: update/execute
+						O: others
+						-- session in transaction --
+						B: begin
+						E: commit/end
+						R: rollback
+						-- session in set session authorization --
+						a: session authorization begin
+						b: session authorization end
+						-- session in copy --
+						C: copy
+						d: copy data
+						e: copy data end
+						-- session in large object --
+						C: create lo
+						O: open lo
+						W: write lo
+						S: lseek lo
+						X: close lo
+						U: unlink lo
+					-- with system call -- 
+						c: confrim commit done
+						q: confrim query done
+						d: detect deadlock
+						x: front end closed
+					-- with cascade connection -- 
+						A: add cascade
+						D: delete cascade
+						F: update all
 					*/
 	char rlog;		/*
 					-- kind of replication log --
@@ -228,8 +311,6 @@ typedef struct ReplicateHeaderType
 						2: send from cluster db (should be retry)
 						3: connection suspended
 					*/
-
-	char dummy;
 	uint16_t port;
 	uint16_t pid;
 	uint32_t query_size;
@@ -238,7 +319,6 @@ typedef struct ReplicateHeaderType
 	char userName[USERNAME_MAX_LENGTH];
 	struct timeval tv;
 	uint32_t query_id;
-    int isAutoCommit; /* 0 if autocommit is off. 1 if autocommit is on */
 	uint32_t request_id;
 	uint32_t replicate_id;
 } ReplicateHeader;
@@ -274,7 +354,6 @@ typedef struct
 	char * next;
 } ConfDataType;
 
-
 typedef struct ReplicateServerInfoType
 {
 	uint32_t useFlag;
@@ -289,7 +368,6 @@ typedef struct ReplicateServerInfoType
 	uint16_t response_mode;
 } ReplicateServerInfo;
 
-
 typedef struct ReplicateNowType
 {
 	int useFlag;
@@ -300,25 +378,89 @@ typedef struct ReplicateNowType
 	struct timeval tp;
 } ReplicateNow;
 
-typedef struct CopyDataType
-{
-	int cnt;
-	char copy_data[COPYBUFSIZ];
-} CopyData;
-
 typedef struct ClusterDBInfoType
 {
 	int status;
 } ClusterDBInfo;
 
+typedef struct
+{
+	uint32_t arg1;
+	uint32_t arg2;
+	uint32_t arg3;
+	char * buf;
+} LOArgs;
 
-extern ConfDataType * ConfData_Top;
-extern ConfDataType * ConfData_End;
-extern ReplicateServerInfo * ReplicateServerData;
-extern ClusterDBInfo * ClusterDBData;
-extern int ReplicateServerShmid;
-extern int ClusterDBShmid;
+/*
+ * cluster server table
+ */
+typedef struct {
+	uint32_t useFlag;
+	char hostName[HOSTNAME_MAX_LENGTH];
+	uint16_t port;
+	uint16_t recoveryPort;
+	uint16_t lifecheckPort;
+	uint32_t hostNum;
+}HostTbl;
 
+/*
+ * host table for recovery request and
+ * load balancer table
+ */
+typedef struct {
+	uint32_t useFlag;
+	char hostName[HOSTNAME_MAX_LENGTH];
+	uint16_t port;
+	uint16_t recoveryPort;
+	uint16_t lifecheckPort;
+	int sock;
+	int recovery_sock;
+} RecoveryTbl;
+
+typedef struct MapTableType
+{
+	uint32_t *useFlag;
+	char *hostName;
+	uint16_t  *port;
+	uint16_t  *lifecheckPort;
+	int sock;
+	int status;
+	int rec_no;
+	pid_t pid;
+} MapTable;
+
+/* for watch dog */
+typedef struct {
+	int SynSemID;
+	int MapTableShmid;
+	int LifeCheckStartShmid;
+	char * LifeCheckStartFlag;
+	MapTable * Map_Table;
+	int LifeCheckTimeOut;
+}PGR_WatchDog_Info;
+
+typedef struct {
+	ConfDataType * ConfData_Top;
+	ConfDataType * ConfData_End;
+	uint16_t LifeCheck_Port_Number;
+	uint16_t Recovery_Port_Number;
+}PGR_ConfData_Info;
+
+/* for logs */
+typedef struct {
+	FILE * StatusFp;
+	FILE * LogFp;
+	char * PGRStatusFileName;
+	char * PGRLogFileName;
+	int Log_Print;
+	int Debug_Print;
+}PGR_Log_Info;
+
+/* PGCluster global */
+typedef struct {
+	PGR_ConfData_Info ConfData_Info;
+	PGR_Log_Info Log_Info;
+}PGR_Com_Info;
 
 /* in backend/libpq/replicate_com.c */
 extern int PGR_Create_Socket_Connect(int * fdP, char * hostName , unsigned short portNumber);
@@ -329,10 +471,24 @@ extern int PGR_Free_Conf_Data(void);
 extern int PGR_Get_Conf_Data(char * dir , char * fname);
 extern void PGRset_recovery_packet_no(RecoveryPacket * packet, int packet_no);
 extern unsigned int PGRget_ip_by_name(char * host);
+extern void PGRsem_unlock( int semid, int sem_num );
+extern void PGRsem_lock( int semid, int sem_num );
 extern void PGRwrite_log_file(FILE * fp, const char * fmt,...);
 extern void show_debug(const char * fmt,...);
 extern void show_error(const char * fmt,...);
+extern void PGR_Init_Com_Info(PGR_Com_Info * com_info);
 
-
+/* in backend/libpq/watch_dog.c */
+extern int PGRsyn_init(void);
+extern void PGRsyn_quit(void);
+extern int PGRmap_to_cluster_tbl(HostTbl * ptr);
+extern int PGRmap_to_replication_server_tbl(ReplicateServerInfo * ptr);
+extern int PGRmap_to_load_balancer_tbl(RecoveryTbl * ptr);
+extern pid_t PGR_ACK_Main(uint16_t port);
+extern pid_t PGR_SYN_Main(int syn_target, void * status_tbl);
+extern void PGRset_start_life_check(void);
+extern void PGRexit_life_check(int sig);
+extern void PGRquit_lifecheck_child(int sig);
+extern void PGRchild_wait(int sig);
 
 #endif /* REPLICATE_COM_H */
