@@ -23,6 +23,8 @@
  * is" without express or implied warranty.
  *
 */
+#ifdef USE_REPLICATION
+
 #include "postgres.h"
 
 #include <stdio.h>
@@ -30,7 +32,6 @@
 #include <unistd.h>
 #include <signal.h>
 #include <sys/wait.h>
-#include <sys/time.h>
 #include <ctype.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -40,6 +41,7 @@
 #include <netinet/in.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <sys/time.h>
 #include <time.h>
 #include <sys/param.h>
 #include <arpa/inet.h>
@@ -80,7 +82,6 @@ int PGRset_status_to_child_tbl(pid_t pid, int status);
 int PGRadd_child_tbl(ClusterTbl * cluster_p, pid_t pid, int status);
 int PGRget_child_status(pid_t pid);
 void PGRreturn_connection_full_error(void);
-void PGRreturn_no_connection_error(void);
 void PGRquit_children_on_cluster(int rec_no);
 
 #ifdef NONE_BLOCK
@@ -94,8 +95,6 @@ static void cancel_request(CancelPacket *sp, int secondary_backend);
 static POOL_CONNECTION_POOL *connect_backend(PGR_StartupPacket *sp, POOL_CONNECTION *frontend);
 static int send_params(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 static void child_end(int sig);
-static void PGRreturn_with_error(char *msg);
-
 
 /*--------------------------------------------------------------------
  * SYMBOL
@@ -184,6 +183,7 @@ PGRdo_child( int use_pool)
 	int connection_reuse = 1;
 	int ssl_request = 0;
 	int count = 0;
+	ClusterTbl * cluster_p = NULL;
 
 	pid = getpid();
 #ifdef PRINT_DEBUG
@@ -191,7 +191,7 @@ PGRdo_child( int use_pool)
 #endif			
 
 	/* set up signal handlers */
-	signal(SIGALRM, SIG_DFL);
+	signal(SIGALRM, SIG_IGN);
 	signal(SIGTERM, child_end);
 	signal(SIGHUP, child_end);
 	signal(SIGINT, child_end);
@@ -204,13 +204,12 @@ PGRdo_child( int use_pool)
 	set_nonblock(Frontend_FD.inet_fd);
 #endif
 
-retry_accept:
 	/* perform accept() */
 	frontend = do_accept(Frontend_FD.unix_fd,Frontend_FD.inet_fd);
 	if (frontend == NULL)
 	{
 		/* accept() failed. return to the accept() loop */
-		PGRset_status_to_child_tbl(pid,TBL_FREE);
+		PGRset_status_to_child_tbl(pid,TBL_ERROR);
 		return STATUS_ERROR;
 	}
 
@@ -234,7 +233,7 @@ retry_startup:
 		/* failed to read the startup packet. return to the
 		   accept() loop */
 		pool_close(frontend);
-		PGRset_status_to_child_tbl(pid,TBL_FREE);
+		PGRset_status_to_child_tbl(pid,TBL_ERROR);
 		return STATUS_ERROR;
 	}
 	PGRset_status_to_child_tbl(pid,TBL_ACCEPT);
@@ -281,13 +280,24 @@ retry_startup:
 	{
 		connection_reuse = 0;
 
+reconnect_backend:
 		if ((backend = connect_backend(sp, frontend)) == NULL)
 		{
-			/*
+			if (use_pool)
+			{
+				show_error("under use_pool return error");
+				return STATUS_ERROR;
+			}
 			PGRset_status_on_cluster_tbl(TBL_ERROR,CurrentCluster);
-			return STATUS_ERROR;
-			*/
-			goto retry_accept;
+			cluster_p = PGRscan_cluster();
+			if (cluster_p == NULL)
+			{
+				show_error("there is no usable clusterdb");
+				return STATUS_ERROR;
+			}
+			CurrentCluster = cluster_p;
+			pool_clear();
+			goto reconnect_backend;
 		}
 	}
 	else
@@ -344,7 +354,7 @@ retry_startup:
 			case POOL_END:
 				/* do not cache connection to template0, template1, regression */
 				if (!strcmp(sp->database, "template0") || !strcmp(sp->database, "template1") ||
-				    !strcmp(sp->database, "regression") || use_pool == NOT_USE_CONNECTION_POOL)
+					!strcmp(sp->database, "regression"))
 				{
 					pool_close(frontend);
 					pool_send_frontend_exits(backend);
@@ -376,7 +386,6 @@ retry_startup:
 
 			/* fatal error occured. just exit myself... */
 			case POOL_FATAL:
-				show_error("%s:do_child: fatal error occured",func);
 				notice_backend_error();
 				break;
 
@@ -398,7 +407,7 @@ retry_startup:
 	if ((status == POOL_ERROR) || 
 		(status == POOL_FATAL))
 	{
-		PGRset_status_to_child_tbl(pid,TBL_FREE);
+		PGRset_status_to_child_tbl(pid,TBL_ERROR);
 		return STATUS_ERROR;
 	}
 	PGRset_status_to_child_tbl(pid,TBL_INIT);
@@ -521,14 +530,12 @@ void
 notice_backend_error(void)
 {
 	pid_t pid = getpid();
-
+	pid_t parent = getppid();
 	PGRset_status_to_child_tbl(pid,TBL_ERROR);
 	PGRset_status_on_cluster_tbl(TBL_ERROR_NOTICE,CurrentCluster);
 
-	/*
 	kill(parent, SIGUSR2);
 	sleep(1);
-	*/
 }
 
 
@@ -540,16 +547,20 @@ do_pooling_child(int sig)
 {
 	char * func = "do_pooling_child()";
 	int rtn;
-	pid_t pid;
 
+#ifdef PRINT_DEBUG
+	pid_t pid;
 	pid = getpid();
+	show_debug("%s:pid[%d] called",func,pid);
+#endif
+
 	rtn = PGRdo_child(USE_CONNECTION_POOL);
-	PGRrelease_connection(CurrentCluster);
 	if (rtn != STATUS_OK)
 	{
 		show_error("%s:PGRdo_child failed",func);
-		child_end(SIGTERM);
+		notice_backend_error();
 	}
+	PGRrelease_connection(CurrentCluster);
 	return ;
 }
 
@@ -641,21 +652,10 @@ PGRget_child_status(pid_t pid)
 
 void 
 PGRreturn_connection_full_error(void)
-{
-  PGRreturn_with_error( "Sorry, backend connection is full\n");
-}
-
-void 
-PGRreturn_no_connection_error(void) {
-  PGRreturn_with_error("pglb could not connect to server: no cluster available.\n");
-}
-
-static void 
-PGRreturn_with_error (char *msg) 
 {
 	PGR_StartupPacket *sp = NULL;
 	POOL_CONNECTION *frontend = NULL;
-
+	static char *msg = "Sorry, backend connection is full";
 
 	/* perform accept() */
 	frontend = do_accept(Frontend_FD.unix_fd,Frontend_FD.inet_fd);
@@ -901,7 +901,6 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	/* read startup packet length */
 	if (pool_read(cp, &len, sizeof(len)))
 	{
-		free(sp);
 		return NULL;
 	}
 	len = ntohl(len);
@@ -910,22 +909,18 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	if (len <= 0)
 	{
 		show_error("%s:read_startup_packet: incorrect packet length (%d)", func,len);
-		free(sp);
-		return NULL;
 	}
 
 	sp->startup_packet = calloc(len, 1);
 	if (!sp->startup_packet)
 	{
 		show_error("%s:read_startup_packet: out of memory",func);
-		free(sp);
 		return NULL;
 	}
 
 	/* read startup packet */
 	if (pool_read(cp, sp->startup_packet, len))
 	{
-		free(sp);
 		return NULL;
 	}
 
@@ -944,7 +939,6 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			if (!sp->database)
 			{
 				show_error("%s:read_startup_packet: out of memory",func);
-				free(sp);
 				return NULL;
 			}
 			strncpy(sp->database, sp2->database, SM_DATABASE);
@@ -953,7 +947,6 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			if (!sp->user)
 			{
 				show_error("%s:read_startup_packet: out of memory",func);
-				free(sp);
 				return NULL;
 			}
 			strncpy(sp->user, sp2->user, SM_USER);
@@ -972,7 +965,6 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 					if (!sp->user)
 					{
 						show_error("%s:read_startup_packet: out of memory",func);
-						free(sp);
 						return NULL;
 					}
 				}
@@ -983,7 +975,6 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 					if (!sp->database)
 					{
 						show_error("%s:read_startup_packet: out of memory",func);
-						free(sp);
 						return NULL;
 					}
 				}
@@ -997,21 +988,18 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			if (!sp->database)
 			{
 				show_error("%s:read_startup_packet: out of memory",func);
-				free(sp);
 				return NULL;
 			}
 			sp->user = calloc(1, 1);
 			if (!sp->user)
 			{
 				show_error("%s:read_startup_packet: out of memory",func);
-				free(sp);
 				return NULL;
 			}
 			break;
 
 		default:
 			show_error("%s:read_startup_packet: invalid major no: %d",func, sp->major);
-			free(sp);
 			return NULL;
 	}
 
@@ -1094,7 +1082,7 @@ static void cancel_request(CancelPacket *sp, int secondary_backend)
 
 static POOL_CONNECTION_POOL *connect_backend(PGR_StartupPacket *sp, POOL_CONNECTION *frontend)
 {
-	char * func ="connect_backend()";
+	char * func = "connect_backend()";
 	POOL_CONNECTION_POOL *backend;
 
 	/* connect to the backend */
@@ -1192,3 +1180,6 @@ child_end(int sig)
 	pool_finish();
 	exit(0);
 }
+
+
+#endif /* USE_REPLICATION */
