@@ -22,14 +22,16 @@
  * is" without express or implied warranty.
  *
 */
+#ifdef USE_REPLICATION
+
 #include "postgres.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 #include <signal.h>
 #include <sys/wait.h>
-#include <sys/time.h>
 #include <ctype.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -40,6 +42,7 @@
 #include <netinet/in.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <sys/time.h>
 #include <time.h>
 #include <sys/param.h>
 #include <sys/select.h>
@@ -125,6 +128,7 @@ PGRload_balance(void)
 		PGRdo_child(NOT_USE_CONNECTION_POOL );
 		PGRrelease_connection(cluster_p);
 		PGRset_status_to_child_tbl(getpid(), TBL_FREE);
+		pool_finish();
 		exit(0);
 	}
 	else if (pid > 0)
@@ -178,7 +182,7 @@ PGRload_balance_with_pool(void)
 		if ( count > PGLB_CONNECT_RETRY_TIME)
 		{
 			show_error("%s:no cluster available",func);
-       			PGRreturn_no_connection_error();
+			PGRreturn_connection_full_error();
 			return STATUS_ERROR;
 		}
 		cluster_p = PGRscan_cluster();
@@ -235,18 +239,4 @@ PGRrelease_connection(ClusterTbl * ptr)
 	}
 	PGRsem_unlock(ClusterSemid,MAX_DB_SERVER);
 }
-
-void
-PGRchild_wait(int sig)
-{
-	pid_t pid = 0;
-	int ret = 0;
-
-	do {
-		pid = waitpid(-1,&ret,WNOHANG);
-		if ((pid <= 0) && (WTERMSIG(ret) > 0))
-		{
-			pid = 1;
-		}
-	} while(pid > 0);
-}
+#endif /* USE_REPLICATION */
