@@ -2,15 +2,15 @@
  * FILE:
  *     pglb.h
  *
- * Portions Copyright (c) 2003-2006  Atsushi Mitani
+ * Portions Copyright (c) 2003-2005  Atsushi Mitani
  *--------------------------------------------------------------------
  */
 #ifndef PGLB_H
 #define PGLB_H
 
-#define PGLB_VERSION	"1.0.10"
+#define PGLB_VERSION	"1.1.0b"
 
-#include "../libpgc/libpgc.h"
+extern PGR_Com_Info Com_Info;
 
 /*
  * from pool.h
@@ -326,7 +326,6 @@ typedef struct {
 #define USE_CONNECTION_POOL	(1)
 
 extern int Recv_Port_Number;
-extern int Recovery_Port_Number;
 extern uint16_t LifeCheck_Port_Number;
 extern int Use_Connection_Pool;
 extern int Max_Pool;
@@ -345,7 +344,6 @@ extern char * PGR_Data_Path;
 extern char * PGR_Write_Path;
 extern char * Backend_Socket_Dir;
 extern FrontSocket Frontend_FD;
-extern FILE * StatusFp;
 
 /* for child.c */
 extern POOL_CONNECTION * Frontend;
@@ -370,11 +368,9 @@ extern int PGRset_status_to_child_tbl(pid_t pid, int status);
 extern int PGRadd_child_tbl(ClusterTbl * cluster_p, pid_t pid, int status);
 extern int PGRget_child_status(pid_t pid);
 extern void PGRreturn_connection_full_error(void);
-extern void PGRreturn_no_connection_error(void);
 extern void PGRquit_children_on_cluster(int rec_no);
 
 /* extern of cluster_table.c */
-extern int PGRis_cluster_alive(void) ;
 extern ClusterTbl * PGRscan_cluster(void);
 extern void PGRset_key_of_cluster(ClusterTbl * ptr, RecoveryPacket * packet);
 extern ClusterTbl * PGRadd_cluster_tbl (ClusterTbl * conf_data);
@@ -387,7 +383,6 @@ extern int PGRload_balance_with_pool(void);
 extern char PGRis_connection_full(ClusterTbl * ptr);
 extern void PGRuse_connection(ClusterTbl * ptr);
 extern void PGRrelease_connection(ClusterTbl * ptr);
-extern void PGRchild_wait(int sig);
 
 /* extern of recovery.c */
 extern void PGRrecovery_main(void);
@@ -417,6 +412,7 @@ extern int connect_inet_domain_socket(int secondary_backend);
 extern int connect_unix_domain_socket(int secondary_backend);
 extern char PGRis_same_host(char * host1, char * host2);
 extern void pool_finish(void);
+extern void pool_clear(void);
 
 /* extern of pool_process_query.c */
 extern POOL_STATUS pool_process_query(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend, int connection_reuse);
@@ -448,11 +444,4 @@ extern int pool_flush(POOL_CONNECTION *cp);
 extern int pool_write_and_flush(POOL_CONNECTION *cp, void *buf, int len);
 extern char *pool_read_string(POOL_CONNECTION *cp, int *len, int line);
 
-/*
- * external prototype in show.c
- */
-extern void show_error(const char * fmt,...);
-extern void show_debug(const char * fmt,...);
-extern void PGRwrite_log_file(FILE * fp, const char * fmt,...);
-
 #endif /* PGLB_H */
