@@ -2,16 +2,18 @@
  * FILE:
  *     pgreplicate.h
  *
- * Portions Copyright (c) 2003-2006, Atsushi Mitani
+ * Portions Copyright (c) 2003-2005, Atsushi Mitani
  *--------------------------------------------------------------------
  */
 #ifndef PGREPLICATE_H
 #define PGREPLICATE_H
 
-#define PGREPLICATE_VERSION		"1.0.10"
+#define PGREPLICATE_VERSION		"1.1.0b"
 
-#include "lib/dllist.h"
-#include "../libpgc/libpgc.h"
+#if 0
+#define COMMIT_ABORT_TEST	(1)
+#define INSERT_ABORT_TEST	(1)
+#endif
 
 /* cascade packet id */
 #define	CMD_SYS_CASCADE	'C'
@@ -19,7 +21,7 @@
 #define CMD_STS_TO_LOWER 'L'
 #define CMD_TYPE_ADD 'A'
 #define CMD_TYPE_DELTE 'D'
-#define CMD_TYPE_UPDATE_ALL 'A'
+#define CMD_TYPE_UPDATE_ALL 'F'
 
 /* log packet id */
 #define	CMD_SYS_LOG	'L'
@@ -28,20 +30,14 @@
 #define CMD_STS_UPDATE_QUERY 'r'
 #define CMD_STS_UPDATE_TRANSACTION 'u'
 
-#define INIT_TRANSACTION_TBL_NUM (12)
 #define MAX_HOST_NAME_LEN	(128)
 #define MAX_DB_SERVER	(32)
+#define PGR_MAX_SOCKET_QUEUE (10000)
 #define	MAX_CONNECTIONS	(128)
 #define PGR_MAX_TICKETS (0x7FFFFFFF)
 #define PGR_MAX_QUERY_ID (0x7FFFFFFF)
 #define PGR_CONNECT_RETRY_TIME  (3)
 #define PGR_EXEC_RETRY_TIME  (5)
-#define DB_TBL_FREE	(0)
-#define DB_TBL_INIT	(1)
-#define DB_TBL_USE	(2)
-#define DB_TBL_ERROR	(-1)
-#define DB_TBL_TOP	(10)
-#define DB_TBL_END	(11)
 #define RECOVERY_HEADER_MTYPE	(1)
 #define QUERY_LOG_MTYPE (2)
 #define PGREPLICATE_CONF_FILE	"pgreplicate.conf"
@@ -66,6 +62,7 @@
 #define SEM_NUM_OF_RECOVERY_QUEUE	(2)
 /* semapho numner of lock tickets */
 #define SEM_NUM_OF_LOCK	(1)
+
 #define STATUS_LOCK_CONFLICT (2)
 #define STATUS_DEADLOCK_DETECT (3)
 #define STATUS_ABORTED (4)
@@ -93,8 +90,8 @@
 #define PGR_SEND_TIMEOUT	(600)
 #define PGR_LOCK_WAIT_MSEC	(50)
 #define PGR_LOCK_TIMEOUT	(60)
-#define PGR_RECOVERY_RETRY_CNT	(6000)
-#define PGR_RECOVERY_WAIT_MSEC	(500)
+#define PGR_RECOVERY_RETRY_CNT	(60)
+#define PGR_RECOVERY_WAIT_MSEC	(50)
 
 #define PGR_RECOVERY_1ST_STAGE	(1)
 #define PGR_RECOVERY_2ND_STAGE	(2)
@@ -107,7 +104,6 @@
  */
 typedef struct {
 	int useFlag;
-	int lock;
 	unsigned short port;
 	unsigned short pid;
 	unsigned int hostIP;
@@ -115,42 +111,12 @@ typedef struct {
 	char host[HOSTNAME_MAX_LENGTH];
 	char srcHost[HOSTNAME_MAX_LENGTH];
 	char dbName[DBNAME_MAX_LENGTH];
+	char userName[USERNAME_MAX_LENGTH];
 	PGconn	* conn;
-	bool in_transaction;
-	bool exec_copy;
+	char * last;
+	char * next;
 }TransactionTbl;
 
-/*
- * cluster server table
- */
-typedef struct {
-	int useFlag;
-	char hostName[HOSTNAME_MAX_LENGTH];
-	char resolvedName[21];
-	int port;
-	int recoveryPort;
-	int hostNum;
-	int transaction_count;
-}HostTbl;
-
-
-typedef struct {
-	FILE * queue_fp;
-	int current_queue_no;
-} RecoveryQueueInf;
-
-
-/*
- * host table for recovery request
- */
-typedef struct {
-	char hostName[HOSTNAME_MAX_LENGTH];
-	int port;
-	int recoveryPort;
-	int sock;
-	int recovery_sock;
-} RecoveryTbl;
-
 /*
  * status table for recovery
  */
@@ -165,16 +131,14 @@ typedef struct {
 } RecoveryStatusInf;
 
 typedef struct {
-	long mtype;
-	long query_mtype;
-	char mdata[1];
-} RecoveryQueueHeader;
+	FILE * queue_fp;
+	int current_queue_no;
+} RecoveryQueueInf;
 
 typedef struct {
 	long mtype;
-        unsigned int replicationId;
 	char mdata[1];
-} RecoveryQueueQuery;
+} MsgData;
 
 typedef struct {
 	unsigned int entry_ticket;
@@ -239,11 +203,11 @@ typedef struct {
 	char * query;
 	int dest;
 	HostTbl * host_ptr;
-  int current_cluster;
-  TransactionTbl *transaction_tbl;
 } ThreadArgInf;
 
 
+extern PGR_Com_Info Com_Info;
+
 /* replication server data */
 extern uint16_t Port_Number;
 extern uint16_t LifeCheck_Port_Number;
@@ -252,8 +216,10 @@ extern int Reserved_Connections;
 
 /* global tables */
 extern HostTbl * Host_Tbl_Begin;
-extern Dllist * Transaction_Tbl_Begin;
+extern TransactionTbl * Transaction_Tbl_Begin;
 extern TransactionTbl * Transaction_Tbl_End;
+extern TransactionTbl * PGconn_Tbl_Begin;
+extern TransactionTbl * PGconn_Tbl_End;
 extern RecoveryTbl * LoadBalanceTbl;
 extern RecoveryStatusInf * Recovery_Status_Inf;
 extern LockWaitInf * Lock_Wait_Tbl;
@@ -264,52 +230,47 @@ extern CommitLogInf * Commit_Log_Tbl;
 extern QueryLogType * Query_Log_Top;
 extern QueryLogType * Query_Log_End;
 extern ReplicationLogInf * Replicateion_Log;
+/* IPC's id data */
 extern int RecoveryShmid;
-extern int ReplicateSerializationShmid;
-extern int RecoveryMsgShmid;
-extern int *RecoveryMsgid;
+extern int RecoveryMsgid;
 extern int HostTblShmid;
 extern int LockWaitTblShmid;
+extern int LoadBalanceTblShmid;
 extern int CascadeTblShmid;
 extern int CascadeInfShmid;
 extern int CommitLogShmid;
-extern int MaxBackends;
-extern char * PGR_Result;
+extern int QueryLogMsgid;
+extern int QueryLogAnsMsgid;
+extern int PGconnMsgid;
 extern int SemID;
 extern int RecoverySemID;
-extern int CascadeSemID;
 extern int LockSemID;
-extern int VacuumSemID;
+
+extern int MaxBackends;
+extern char * PGR_Result;
 extern char * PGR_Data_Path;
 extern char * PGR_Write_Path;
-extern FILE * LogFp;
-extern FILE * StatusFp;
-extern int Log_Print;
-extern int Debug_Print;
+
 extern char * Function;
+
 extern int IS_SESSION_AUTHORIZATION;
 extern ResponseInf * PGR_Response_Inf;
 extern bool StartReplication[MAX_DB_SERVER];
 extern bool PGR_Cascade;
 extern bool	PGR_Use_Replication_Log;
 extern bool	PGR_AutoCommit;
-extern unsigned int * PGR_ReplicateSerializationID;
 extern unsigned int * PGR_Send_Query_ID;
 extern unsigned int PGR_Query_ID;
-extern volatile bool exit_processing;
 extern RecoveryQueueInf RecoveryQueue;
-extern int pgreplicate_pid;
-
-extern int ReplicateSock;
 
 /* smart shutdown */
 extern int Idle_Flag;
-extern volatile bool Exit_Request;
+extern bool Exit_Request;
 
 /*
  * external prototype in main.c
  */
-extern void child_wait(SIGNAL_ARGS);
+extern void PGRstop_pgreplicate(int sig);
 
 /*
  * external prototype in conf.c
@@ -319,7 +280,6 @@ extern int PGRset_Conf_Data(char * path);
 /*
  * external prototype in replicate.c
  */
-extern int PGRset_replication_id(uint32_t rid);
 extern bool PGRis_same_host(char * host1, unsigned short port1 , char * host2, unsigned short port2);
 extern HostTbl * PGRadd_HostTbl(HostTbl *  conf_data, int useFlag);
 extern HostTbl * PGRget_master(void);
@@ -328,27 +288,25 @@ extern int PGRget_recovery_status(void);
 extern int PGRcheck_recovered_host(void);
 extern int PGRset_recovered_host(HostTbl * target,int useFlag);
 extern int PGRinit_recovery(void);
-extern void set_exit_processing(int signo);
-extern void PGRreplicate_exit(int exit_status);
-extern int PGRsend_replicate_packet_to_server( HostTbl * host_ptr, ReplicateHeader * header, char *query , char * result,unsigned int replicationId, bool recovery);
+extern void PGRreplicate_exit(int signal_arg);
+extern int PGRsend_replicate_packet_to_server( HostTbl * host_ptr, ReplicateHeader * header, char *query , char * result, int type);
 extern int PGRreplicate_packet_send_each_server( HostTbl * ptr, bool return_response, ReplicateHeader * header, char * query,int dest);
 extern HostTbl * PGRget_HostTbl(char * hostName,int port);
 extern int PGRset_queue(ReplicateHeader * header,char * query);
 extern int PGRset_host_status(HostTbl * host_ptr,int status);
-extern void PGRclear_connections(void);
-extern void PGRsem_unlock( int semid, short sem_num );
-extern void PGRsem_lock( int semid, short sem_num );
+extern void PGRclear_transactions(void);
+extern void PGRclear_connections(int use_flag);
+extern int PGRset_replication_id(uint32_t id);
 extern int PGRdo_replicate(int sock,ReplicateHeader *header, char * query);
-extern int PGRreturn_result(int dest, char * result, int wait);
-extern int PGRreplicate_packet_send( ReplicateHeader * header, char * query,int dest,int recovery_status);
+extern int PGRreturn_result(int dest, int wait);
+extern int PGRreplicate_packet_send( ReplicateHeader * header, char * query,int dest);
 extern char * PGRread_packet(int sock, ReplicateHeader *header);
 extern void PGRnotice_replication_server(char * hostName, unsigned short portNumber,unsigned short recoveryPortNumber, unsigned short lifecheckPortNumber, char * userName);
 extern char * PGRread_query(int sock, ReplicateHeader *header);
-extern int PGRwait_transaction_count_clear(HostTbl *host);
+extern int PGRwait_transaction_count_clear(void);
 extern int PGRsync_oid(ReplicateHeader *header);
 extern unsigned int PGRget_next_query_id(void);
-extern int PGRinit_transaction_table(void);
-extern int replicate_packet_send_internal(ReplicateHeader * header, char * query,int dest,int recovery_status,bool isHeldLock);
+extern int PGRget_pgconn_queue(void);
 
 /*
  * external prototype in recovery.c
@@ -390,6 +348,5 @@ extern int PGRsend_cascade(int sock , ReplicateHeader * header, char * query);
 extern int PGRcascade_main(int sock, ReplicateHeader * header, char * query);
 extern int PGRwait_notice_rlog_done(void);
 extern int PGRsend_notice_rlog_done(int sock);
-extern int PGRsend_notice_quit(void);
 
 #endif /* PGREPLICATE_H */
