@@ -6,9 +6,11 @@
  *     This file is composed of the functions to call with the source
  *     at pgreplicate for the recovery.
  *
- * Portions Copyright (c) 2003-2006, Atsushi Mitani
+ * Portions Copyright (c) 2003, Atsushi Mitani
  *--------------------------------------------------------------------
  */
+#ifdef USE_REPLICATION
+
 #include "postgres.h"
 
 #include <stdio.h>
@@ -40,12 +42,10 @@
 
 #include "miscadmin.h"
 #include "nodes/nodes.h"
-
 #include "libpq-fe.h"
 #include "libpq/libpq-fs.h"
 #include "libpq-int.h"
 #include "fe-auth.h"
-
 #include "access/xact.h"
 #include "replicate_com.h"
 #include "pgreplicate.h"
@@ -67,16 +67,15 @@
 #ifdef MULTIBYTE
 #include "mb/pg_wchar.h"
 #endif
-#include "pgreplicate.h"
 
 
 /*--------------------------------------
  * GLOBAL VARIABLE DECLARATION
  *--------------------------------------
  */
-static RecoveryPacket MasterPacketData;
-static RecoveryTbl Master;
-static RecoveryTbl Target;
+RecoveryPacket MasterPacketData;
+RecoveryTbl Master;
+RecoveryTbl Target;
 
 
 /*--------------------------------------
@@ -94,12 +93,16 @@ static void finish_recovery(void);
 static bool first_setup_recovery(int * sock, RecoveryPacket * packet);
 static bool second_setup_recovery (RecoveryPacket * packet);
 static void pgrecovery_loop(int fd);
+static void show_recovery_packet(RecoveryPacket * packet);
 static int PGRsend_queue(RecoveryTbl * master, RecoveryTbl * target);
 static int send_vacuum(HostTbl *host, char * userName, int stage);
+static int send_recovery_queue( HostTbl * master_ptr, HostTbl * target_ptr, ReplicateHeader *header, char * query);
+static void clear_recovery_queue_file(void);
 
-#ifdef PRINT_DEBUG
-static void show_recovery_packet(RecoveryPacket * packet);
-#endif			
+int PGRsend_load_balance_packet(RecoveryPacket * packet);
+void PGRrecovery_main(void);
+FILE * PGRget_recovery_queue_file_for_write(void);
+FILE * PGRget_recovery_queue_file_for_read(int next);
 
 /*-----------------------------------------------------------
  * SYMBOL
@@ -117,10 +120,8 @@ static void show_recovery_packet(RecoveryPacket * packet);
 static int
 read_packet(int sock,RecoveryPacket * packet)
 {
-#ifdef PRINT_DEBUG
-	char * func = "read_packet()";
-#endif			
 	int r = 0;
+	int cnt = 0;
 	char * read_ptr = NULL;
 	int read_size = 0;
 	int packet_size = 0;
@@ -136,33 +137,40 @@ read_packet(int sock,RecoveryPacket * packet)
 		r = recv(sock,read_ptr + read_size ,packet_size - read_size, MSG_WAITALL);
 		if (r < 0)
 		{
-			if (errno == EINTR || errno == EAGAIN)
+			if (errno == EINTR)
+			{
+				usleep(PGR_RECV_WAIT_MSEC);
 				continue;
+			}
+#ifdef EAGAIN
+			if (errno == EAGAIN)
+			{
+				usleep(PGR_RECV_WAIT_MSEC);
+				continue;
+			}
+#endif /* EAGAIN */
+			if (cnt < PGR_RECV_RETRY_CNT )
+			{
+				cnt ++;
+				usleep(PGR_RECV_WAIT_MSEC);
+				continue;
+			}
 			else
 			{
-				show_error("%s:recv failed: %d(%s)",func, errno, strerror(errno));
 				return -1;
 			}
 		}
-		else if (r > 0)
+		if (r > 0)
 		{
 			read_size += r;
-			if (read_size == packet_size)
+			if (read_size == PGR_MESSAGE_BUFSIZE)
 			{
-#ifdef PRINT_DEBUG
-				show_debug("%s:receive packet",func);
 				show_recovery_packet(packet);
-#endif			
 				return read_size;
 			}
 		}
-		else /* r == 0 */
-		{
-			show_error("%s:unexpected EOF", func);
-			return -1;
-		}
+		return -1;
 	}
-	return -1;
 }
 
 static int
@@ -173,23 +181,17 @@ read_packet_from_master( RecoveryTbl * host, RecoveryPacket * packet )
 	fd_set	  rmask;
 	struct timeval timeout;
 
+	timeout.tv_sec = PGR_RECV_TIMEOUT;
+	timeout.tv_usec = 0;
+
+	/*
+	 * Wait for something to happen.
+	 */
+	FD_ZERO(&rmask);
+	FD_SET(host->recovery_sock,&rmask);
 	for(;;)
 	{
-		timeout.tv_sec = PGR_RECV_TIMEOUT;
-		timeout.tv_usec = 0;
-
-		/*
-		 * Wait for something to happen.
-		 */
-		FD_ZERO(&rmask);
-		FD_SET(host->recovery_sock,&rmask);
 		rtn = select(host->recovery_sock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
-		
-		if (rtn == 0) /* timeout */
-		{
-			return -1;
-		}
-
 		if (rtn && FD_ISSET(host->recovery_sock, &rmask))
 		{
 			read_size = read_packet(host->recovery_sock, packet);
@@ -201,35 +203,32 @@ read_packet_from_master( RecoveryTbl * host, RecoveryPacket * packet )
 static int
 send_recovery_packet(int  sock, RecoveryPacket * packet)
 {
-	char *func = "send_recovery_packet";
+	char * func = "send_recovery_packet()";
 	char * send_ptr;
 	int send_size= 0;
 	int buf_size = 0;
 	int s;
 	
+	errno = 0;
 	send_ptr = (char *)packet;
 	buf_size = sizeof(RecoveryPacket);
 
 	for (;;)
 	{
 		s = send(sock, send_ptr + send_size,buf_size - send_size ,0);
-		if (s < 0)
-		{
-			if (errno == EINTR || errno == EAGAIN)
+		if (s < 0){
+			if (errno == EINTR)
+			{
 				continue;
-
-			show_error("%s:send error: %d(%s)", func, errno, strerror(errno));
+			}
+			show_error("%s:send() failed. (%s)",func,strerror(errno));
 			return STATUS_ERROR;
 		}
-		else if (s == 0)
+		if (s == 0)
 		{
-			show_error("%s:unexpected EOF", func);
-			return STATUS_ERROR;
-		}
-
-		send_size += s;
-		if (send_size == buf_size)
 			return STATUS_OK;
+		}
+		send_size += s;
 	}
 }
 
@@ -245,7 +244,7 @@ send_packet(RecoveryTbl * host, RecoveryPacket * packet )
 		{
 			if (count > MAX_RETRY_TIMES )
 			{
-				show_error("%s:host[%s] port[%d]PGR_Create_Socket_Connect failed",func,host->hostName, host->recoveryPort);
+				show_error("%s:PGR_Create_Socket_Connect failed",func);
 				return STATUS_ERROR;
 			}
 			count ++;
@@ -257,10 +256,6 @@ send_packet(RecoveryTbl * host, RecoveryPacket * packet )
 		close(host->recovery_sock);
 		host->recovery_sock = -1;
 		PGR_Create_Socket_Connect(&(host->recovery_sock), host->hostName , host->recoveryPort);
-#ifdef PRINT_DEBUG
-		show_debug("%s:PGR_Create_Socket_Connectsock[%d] host[%s] port[%d]",
-			func,host->recovery_sock,host->hostName,host->recoveryPort);
-#endif
 		if (count > PGR_CONNECT_RETRY_TIME )
 		{
 
@@ -311,19 +306,14 @@ PGRsend_load_balance_packet(RecoveryPacket * packet)
 	}
 	while (lbp->hostName[0] != 0)
 	{
-		if (lbp->recovery_sock != -1)
-		{
-			close(lbp->recovery_sock);
-			lbp->recovery_sock = -1;
-		}
-#ifdef PRINT_DEBUG
-	show_debug("%s:host[%s] port[%d]",func,lbp->hostName,lbp->recoveryPort);
-#endif
 		status = send_packet(lbp,packet);
-		if (lbp->recovery_sock != -1)
+		if (status == STATUS_OK)
 		{
-			close(lbp->recovery_sock);
-			lbp->recovery_sock = -1;
+			if (lbp->sock != -1)
+			{
+				close(lbp->sock);
+				lbp->sock = -1;
+			}
 		}
 		lbp ++;
 	}
@@ -354,8 +344,7 @@ send_vacuum(HostTbl *host, char * userName, int stage)
 	header.cmdType = CMD_TYPE_VACUUM;
 	header.pid = getpid();
 	header.query_id = getpid();
-	header.isAutoCommit=1;
-	rtn = PGRsend_replicate_packet_to_server(host,&header,query,PGR_Result,0, true);
+	rtn = PGRsend_replicate_packet_to_server(host,&header,query,PGR_Result, RECOVERY_QUERY_TYPE);
 	if (query !=NULL)
 		free(query);
 	return rtn;	
@@ -371,7 +360,7 @@ first_setup_recovery(int * sock, RecoveryPacket * packet)
 	HostTbl host_tbl;
 	char * userName = NULL;
 
-	memcpy(Target.hostName,packet->hostName,sizeof(Target.hostName));
+	strncpy(Target.hostName,packet->hostName,sizeof(Target.hostName));
 	Target.port = ntohs(packet->port);
 	Target.recoveryPort = ntohs(packet->recoveryPort);
 	Target.sock = *sock;
@@ -388,9 +377,7 @@ first_setup_recovery(int * sock, RecoveryPacket * packet)
 		/*
 		 * recovery process is already running
 		 */
-#ifdef PRINT_DEBUG
-		show_debug("%s:already recovery job runing",func);
-#endif			
+		show_error("%s:already recovery job runing",func);
 		memset(packet,0,sizeof(packet));
 		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_OCCUPIED) ;
 		status = send_packet(&Target,packet);
@@ -400,13 +387,10 @@ first_setup_recovery(int * sock, RecoveryPacket * packet)
 	/*
 	 * add recovery target to host table
 	 */
-#ifdef PRINT_DEBUG
-	show_debug("%s:add recovery target to host table",func);
-#endif			
-	memcpy(host_tbl.hostName,packet->hostName,sizeof(host_tbl.hostName));
+	strncpy(host_tbl.hostName,packet->hostName,sizeof(host_tbl.hostName));
 	host_tbl.port = ntohs(packet->port);
-	PGRset_recovered_host(&host_tbl,DB_TBL_INIT);
-	PGRadd_HostTbl(&host_tbl,DB_TBL_INIT);
+	PGRset_recovered_host(&host_tbl,DATA_INIT);
+	PGRadd_HostTbl(&host_tbl,DATA_INIT);
 	/*
 	 * send prepare recovery to load balancer
 	 */
@@ -416,9 +400,6 @@ first_setup_recovery(int * sock, RecoveryPacket * packet)
 	/*
 	 * set RECOVERY_PGDATA_REQ packet data
 	 */
-#ifdef PRINT_DEBUG
-	show_debug("%s:set RECOVERY_PGDATA_REQ packet data",func);
-#endif			
 	memset(packet,0,sizeof(RecoveryPacket));
 	PGRset_recovery_packet_no(packet, RECOVERY_PGDATA_REQ );
 
@@ -430,31 +411,37 @@ retry_connect_master:
 		 * connection error , master may be down
 		 */
 		show_error("%s:get master info error , master may be down",func);
-		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_TARGET_ONLY);
+		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
 		status = send_packet(&Target, packet);
 		reset_recovery_prepare();
 		loop_end = true;
 		if (userName != NULL)
+		{
 			free(userName);
+		}
 		return loop_end;
 	}
 	/* send vauum command to master server */
 	status = send_vacuum(master, userName, PGR_RECOVERY_1ST_STAGE );
 	if (status != STATUS_OK)
 	{
-		PGRset_host_status(master, DB_TBL_ERROR);
-		goto retry_connect_master;
+		show_error("%s:vacuum error , master may be down",func);
+		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
+		status = send_packet(&Target, packet);
+		reset_recovery_prepare();
+		loop_end = true;
+		if (userName != NULL)
+		{
+			free(userName);
+		}
+		return loop_end;
 	}
 
-	memcpy(Master.hostName,master->hostName,sizeof(Master.hostName));
+	strncpy(Master.hostName,master->hostName,sizeof(Master.hostName));
 	Master.sock = -1;
 	Master.recovery_sock = -1;
 	Master.port = master->port;
 	Master.recoveryPort = master->recoveryPort;
-
-#ifdef PRINT_DEBUG
-	show_debug("%s:send packet to master %s recoveryPort %d",func, Master.hostName, Master.recoveryPort);
-#endif			
 	status = send_packet(&Master, packet);
 	if (status != STATUS_OK)
 	{
@@ -462,7 +449,7 @@ retry_connect_master:
 		 * connection error , master may be down
 		 */
 		show_error("%s:connection error , master may be down",func);
-		PGRset_host_status(master,DB_TBL_ERROR);
+		PGRset_host_status(master,DATA_ERR);
 		goto retry_connect_master ;
 	}
 	
@@ -475,14 +462,8 @@ retry_connect_master:
 	/*
 	 * wait answer from master server 
 	 */
-#ifdef PRINT_DEBUG
-	show_debug("%s:wait answer from master server",func);
-#endif			
 	memset(packet,0,sizeof(RecoveryPacket));
 	read_packet_from_master(&Master, packet);
-#ifdef PRINT_DEBUG
-	show_debug("%s:get answer from master:no[%d]",func,ntohs(packet->packet_no));
-#endif			
 	if (ntohs(packet->packet_no) == RECOVERY_PGDATA_ANS)
 	{
 		/*
@@ -496,12 +477,11 @@ retry_connect_master:
 		 * prepare answer from master DB
 		 */
 		PGRset_recovery_packet_no(packet, RECOVERY_PREPARE_ANS );
-		memcpy(packet->hostName,Master.hostName,sizeof(packet->hostName));
+		strncpy(packet->hostName,Master.hostName,sizeof(packet->hostName));
 		status = send_packet(&Target, packet);
 		if (status != STATUS_OK)
 		{
-			show_error("%s:no[%d] send_packet to target error",func,ntohs(packet->packet_no));
-			PGRset_recovery_packet_no(packet, RECOVERY_ERROR_TARGET_ONLY);
+			PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
 			status = send_packet(&Master,packet);
 			reset_recovery_prepare();
 			loop_end = true;
@@ -509,8 +489,6 @@ retry_connect_master:
 	}
 	if (userName != NULL)
 		free(userName);
-
-
 	return loop_end;
 }
 
@@ -524,59 +502,36 @@ second_setup_recovery (RecoveryPacket * packet)
 	char * userName = NULL;
 
 	start_recovery();
-
-	/* send vauum command to master server */
-	while ((master = PGRget_master()) != NULL)
-	{
-		userName = strdup(packet->userName);
-		status = send_vacuum(master, userName, PGR_RECOVERY_2ND_STAGE );
-		if (status != STATUS_OK)
-		{
-			PGRset_host_status(master, DB_TBL_ERROR);
-			if (userName != NULL)
-			{
-				free(userName);
-				userName = NULL;
-			}
-			continue;
-		}
-		break;
-	}
-
-	if (master == NULL)
-	{
-		show_error("%s:vacuum error , master may be down",func);
-		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_TARGET_ONLY);
-		status = send_packet(&Target,packet);
-		status = send_packet(&Master,packet);
-		reset_recovery_prepare();
-
-		return true;		
-	}
-
 	/*
 	 * wait until all started transactions are going to finish
 	 */
-	status = PGRwait_transaction_count_clear(master);
-	if (status != STATUS_OK && PGRget_recovery_status() == RECOVERY_START)
+	status = PGRwait_transaction_count_clear();
+	if (status != STATUS_OK)
 	{
 		show_error("%s:transaction is too busy, please try again after",func);
-		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_TARGET_ONLY);
-		status = send_packet(&Target,packet);
-		status = send_packet(&Master,packet);
-		reset_recovery_prepare();
-		return true;
-	}
-
-	if (PGRget_recovery_status() != RECOVERY_START) {
-		show_error("%s:queue set failed. stop to recovery",func);
 		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
 		status = send_packet(&Target,packet);
 		status = send_packet(&Master,packet);
 		reset_recovery_prepare();
+		loop_end = true;
+		return loop_end;
+	}
+
+	/* send vauum command to master server */
+	master = PGRget_master();
+	userName = strdup(packet->userName);
+	status = send_vacuum(master, userName, PGR_RECOVERY_2ND_STAGE );
+	if (status != STATUS_OK)
+	{
+		show_error("%s:vacuum error , master may be down",func);
+		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
+		status = send_packet(&Target,packet);
+		status = send_packet(&Master,packet);
+		reset_recovery_prepare();
+		loop_end = true;
 		if (userName != NULL)
 			free(userName);
-		return true;
+		return loop_end;
 	}
 
 	/*
@@ -594,22 +549,11 @@ second_setup_recovery (RecoveryPacket * packet)
 		status = send_packet(&Target,packet);
 		status = send_packet(&Master,packet);
 		reset_recovery_prepare();
+		loop_end = true;
 		if (userName != NULL)
 			free(userName);
-		return true;
+		return loop_end;
 	}
-
-	if (PGRget_recovery_status() != RECOVERY_START) {
-         	show_error("%s:queue set failed. stop to recovery",func);
-		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
-		status = send_packet(&Target,packet);
-		status = send_packet(&Master,packet);
-		reset_recovery_prepare();
-		if (userName != NULL)
-			free(userName);
-		return true;
-	}
-
 	/*
 	 * wait answer from master server 
 	 */
@@ -621,7 +565,7 @@ second_setup_recovery (RecoveryPacket * packet)
 		 * master DB finished fsync
 		 */
 		PGRset_recovery_packet_no(packet, RECOVERY_START_ANS );
-		memcpy(packet->hostName,Master.hostName,sizeof(packet->hostName));
+		strncpy(packet->hostName,Master.hostName,sizeof(packet->hostName));
 		status = send_packet(&Target,packet);
 		if (status != STATUS_OK)
 		{
@@ -631,7 +575,7 @@ second_setup_recovery (RecoveryPacket * packet)
 	}
 	else
 	{
-		show_error("%s:failure answer returned",func);
+		show_error("%s:failer answer returned",func);
 		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
 		status = send_packet(&Target,packet);
 		status = send_packet(&Master,packet);
@@ -642,36 +586,32 @@ second_setup_recovery (RecoveryPacket * packet)
 		free(userName);
 	return loop_end;
 }
-/**
- * send queries from queue.
- *
- * return
- *   STATUS_OK - success both
- *   STATUS_ERROR - fail both
- */
+
 static int
 PGRsend_queue(RecoveryTbl * master, RecoveryTbl * target)
 {
 	char * func = "PGRsend_queue()";
 	HostTbl * master_ptr = NULL;
 	HostTbl * target_ptr = NULL;
-	RecoveryQueueHeader * msg_header = NULL;
-	RecoveryQueueQuery * msg_query = NULL;
-	ReplicateHeader * header = NULL;
+	ReplicateHeader header;
 	char * query = NULL;
-	int size = 0;
-	int rcv_header_size = 0;
-	long mtype = 0;
 	int status = 0;
 	int query_size = 0;
-	int rtn=0;
+	int alloc_size = 0;
+	int read_lock;
+
+	if (RecoverySemID == 0)
+	{
+		show_error("%s:initial setup should be failed",func);
+		return STATUS_ERROR;
+	}
 	if (master == (RecoveryTbl *)NULL)
 	{
 		show_error("%s:there is no master ",func);
 		return STATUS_ERROR;
 	}
 #ifdef PRINT_DEBUG
-	show_debug("%s:master %s - %d",func,master->hostName,master->port);
+	show_debug("%s:master %s - %d\n",func,master->hostName,master->port);
 #endif			
 	master_ptr = PGRget_HostTbl(master->hostName,master->port);
 	if (master_ptr == (HostTbl *)NULL)
@@ -691,98 +631,110 @@ PGRsend_queue(RecoveryTbl * master, RecoveryTbl * target)
 			return STATUS_ERROR;
 		}
 	}
-
-	size = sizeof(RecoveryQueueHeader) + sizeof(ReplicateHeader);
-	msg_header = (RecoveryQueueHeader *)malloc(size+4);
-	if (msg_header == NULL)
+	RecoveryQueue.queue_fp = PGRget_recovery_queue_file_for_read(0);
+	if (RecoveryQueue.queue_fp == NULL)
 	{
-		show_debug("%s:malloc() failed. reason: %s",func, strerror(errno));
 		return STATUS_ERROR;
 	}
-	memset(msg_header,0,size+4);
-	rcv_header_size = sizeof(long)+sizeof(ReplicateHeader);
+	clearerr(RecoveryQueue.queue_fp);
+	fseek(RecoveryQueue.queue_fp,0,SEEK_SET);
 	status = STATUS_OK;
-	while (msgrcv(*RecoveryMsgid , msg_header, rcv_header_size, RECOVERY_HEADER_MTYPE, IPC_NOWAIT) > 0 )
+	read_lock = 0;
+	while (RecoveryQueue.queue_fp != NULL)
 	{
-		header = (ReplicateHeader *)(msg_header->mdata);
-		query_size = ntohl(header->query_size);
-		if (query_size < 0)
+		if(feof(RecoveryQueue.queue_fp) != 0)
 		{
-			if (msg_query != NULL)
+			RecoveryQueue.queue_fp = PGRget_recovery_queue_file_for_read(1);
+			if (RecoveryQueue.queue_fp == NULL)
 			{
-				free(msg_query);
-				msg_query = NULL;
-			}
-			break;
-		}
-		size = sizeof(RecoveryQueueQuery) + query_size;
-		mtype = msg_header->query_mtype;
-		msg_query = (RecoveryQueueQuery *) malloc(size+4);
-		if (msg_query == NULL)
-		{
-			show_debug("%s:malloc() failed. reason: %s",func, strerror(errno));
-			free(msg_header);
-			msg_header = NULL;
-			return STATUS_ERROR;
-		}
-		memset(msg_query,0,size+4);
-		if (msgrcv(*RecoveryMsgid , msg_query, sizeof(unsigned int)+query_size, mtype, IPC_NOWAIT) < 0)
-		{
-			show_error("%s:msgrcv() failed. (%s)",func,strerror(errno) );
-			status = STATUS_ERROR;
-			if (msg_query != NULL)
-			{
-				free(msg_query);
-				msg_query = NULL;
-			}
-			break;
-		}
-		query = (char *)(msg_query->mdata);
-		show_debug("%s:put query from queue :%s",func,query);
-		rtn=PGRsend_replicate_packet_to_server(master_ptr,header,query,PGR_Result,msg_query->replicationId, true);
-		if (rtn == STATUS_ERROR || rtn == STATUS_DEADLOCK_DETECT )
-		{
-			show_error("%s:PGRsend_replicate_packet_to_server to master error",func);
-			status = STATUS_ERROR ;
-			if (msg_query != NULL)
-			{
-				free(msg_query);
-				msg_query = NULL;
-			}
-			break;
-		}
-		if (target_ptr != NULL)
-		{
-			rtn=PGRsend_replicate_packet_to_server(target_ptr,header,query,PGR_Result,msg_query->replicationId, true);
-			if (rtn == STATUS_ERROR || rtn == STATUS_DEADLOCK_DETECT )
-			{
-				show_error("%s:PGRsend_replicate_packet_to_server to target error",func);
-				status = STATUS_ERROR;
-				if (msg_query != NULL)
-				{
-					free(msg_query);
-					msg_query = NULL;
-				}
+				clear_recovery_queue_file();
 				break;
 			}
+			else
+			{
+				clearerr(RecoveryQueue.queue_fp);
+				fseek(RecoveryQueue.queue_fp,0,SEEK_SET);
+			}
 		}
-		if (msg_query != NULL)
+		if (!read_lock)
 		{
-			free(msg_query);
-			msg_query = NULL;
+			read_lock = 1;
+			PGRsem_lock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+		}
+		if (fread(&header,sizeof(ReplicateHeader),1,RecoveryQueue.queue_fp) >0)
+		{
+			query_size = ntohl(header.query_size);
+			if (query_size < 0)
+			{
+				continue;
+			}
+			else if (query_size > 0)
+			{
+				if (alloc_size < query_size +4)
+				{
+					if (alloc_size == 0)
+					{
+						query = (char *)malloc(query_size+4);
+					}
+					else
+					{
+						query = (char*)realloc(query,query_size+4);
+					}
+					if (query == (char*)NULL)
+					{
+						show_error("%s:malloc orrealloc faild: (%s)",func,strerror(errno));
+						PGRsem_unlock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+						return STATUS_ERROR;
+					}
+				}
+				memset(query,0,query_size+4);
+				if (fread(query,query_size,1, RecoveryQueue.queue_fp) >0)
+				{
+					read_lock = 0;
+					PGRsem_unlock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+					status = send_recovery_queue(master_ptr,target_ptr, &header, query);
+				}
+			}
+		}
+		if (read_lock)
+		{
+			read_lock = 0;
+			PGRsem_unlock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+		}
+		if (ferror(RecoveryQueue.queue_fp) != 0)
+		{
+			status = STATUS_ERROR;
+			break;
 		}
 	}
-	if (msg_header != NULL)
+	if (query != NULL)
 	{
-		free(msg_header);
-		msg_header = NULL;
+		free(query);
 	}
-#ifdef PRINT_DEBUG
-	show_debug("%s:send_queue return status %d",func,status);
-#endif			
 	return status;
 }
 
+static int
+send_recovery_queue( HostTbl * master_ptr, HostTbl * target_ptr, ReplicateHeader *header, char * query)
+{
+	char * func = "send_recovery_queue()";
+
+	if (PGRsend_replicate_packet_to_server(master_ptr,header,query,PGR_Result, RECOVERY_QUERY_TYPE) != STATUS_OK)
+	{
+		show_error("%s:PGRsend_replicate_packet_to_server to master error",func);
+		return STATUS_ERROR;
+	}
+	if (target_ptr != NULL)
+	{
+		if (PGRsend_replicate_packet_to_server(target_ptr,header,query,PGR_Result, RECOVERY_QUERY_TYPE) != STATUS_OK)
+		{
+			show_error("%s:PGRsend_replicate_packet_to_server to target error",func);
+			return STATUS_ERROR;
+		}
+	}
+	return STATUS_OK;
+}
+
 static void
 pgrecovery_loop(int fd)
 {
@@ -793,17 +745,7 @@ pgrecovery_loop(int fd)
 	bool loop_end = false;
 	RecoveryPacket packet;
 	HostTbl new_host;
-	RecoveryTbl * lbp;
 
-	lbp = LoadBalanceTbl;
-	if (lbp == (RecoveryTbl *)NULL)
-	{
-		show_error("%s:recovery table is NULL",func);
-		return ;
-	}
-#ifdef PRINT_DEBUG
-	show_debug("%s:recovery accept port %d",func, Recovery_Port_Number);
-#endif			
 	count = 0;
 	while ((status = PGR_Create_Acception(fd,&sock,"",Recovery_Port_Number)) != STATUS_OK)
 	{
@@ -816,10 +758,6 @@ pgrecovery_loop(int fd)
 		}
 		count ++;
 	}
-	if(sock==-1) {
-			show_error("can't create recovery socket.exit.");
-			PGRreplicate_exit(1);
-	}
 	for(;;)
 	{
 		int read_size = 0;
@@ -839,12 +777,6 @@ pgrecovery_loop(int fd)
 		 * read packet from target cluster server
 		 */
 		rtn = select(sock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
-
-		if (rtn == 0) /* timeout */
-		{
-			return;
-		}
-
 		if (rtn && FD_ISSET(sock, &rmask))
 		{
 			read_size = read_packet(sock, &packet);
@@ -864,16 +796,13 @@ pgrecovery_loop(int fd)
 				/*
 				 * start prepare of recovery
 				 */
-
+				loop_end = first_setup_recovery(&sock, &packet);
 #ifdef PRINT_DEBUG
 				show_debug("%s:1st master %s - %d",
 					func,Master.hostName,Master.port);
 				show_debug("%s:1st target %s - %d",
 					func,Target.hostName,Target.port);
 #endif			
-
-				loop_end = first_setup_recovery(&sock, &packet);
-				show_debug("%s:first_setup_recovery end:%d ",func,loop_end);
 				break;
 			case RECOVERY_START_REQ : 
 				/*
@@ -881,14 +810,12 @@ pgrecovery_loop(int fd)
 				 *    stop the transaction count up
 				 *    start queueing and stop send all queries for master DB
 				 */
-#ifdef PRINT_DEBUG
-				show_debug("%s:2nd master %s - %d",
-					func, Master.hostName,Master.port);
-				show_debug("%s:2nd target %s - %d",
-					func, Target.hostName,Target.port);
-#endif			
 				loop_end = second_setup_recovery (&packet);
 #ifdef PRINT_DEBUG
+				show_debug("%s:2nd master %s - %d",
+					func,Master.hostName,Master.port);
+				show_debug("%s:2nd target %s - %d",
+					func,Target.hostName,Target.port);
 				show_debug("%s:second_setup_recovery end :%d ",
 					func,loop_end);
 #endif			
@@ -900,21 +827,24 @@ pgrecovery_loop(int fd)
 				/*
 				 * send all queries in queue
 				 */
-
 #ifdef PRINT_DEBUG
 				show_debug("%s:last master %s - %d",
-					func, Master.hostName,Master.port);
+					func,Master.hostName,Master.port);
 				show_debug("%s:last target %s - %d",
-					func, Target.hostName,Target.port);
+					func,Target.hostName,Target.port);
 #endif			
+				sleep(5);
 				status = PGRsend_queue(&Master,&Target);
 				if (status == STATUS_OK)
 				{
-					memcpy(new_host.hostName,Target.hostName,sizeof(new_host.hostName));
+#ifdef PRINT_DEBUG
+					show_debug("%s:PGRsend_queue ok",func);
+#endif			
+					strncpy(new_host.hostName,Target.hostName,sizeof(new_host.hostName));
 					new_host.port = Target.port;
 					new_host.recoveryPort = Target.recoveryPort;
-					PGRset_recovered_host(&new_host,DB_TBL_USE);
-					PGRadd_HostTbl(&new_host,DB_TBL_USE);
+					PGRset_recovered_host(&new_host,DATA_USE);
+					PGRadd_HostTbl(&new_host,DATA_USE);
 				}
 				else
 				{
@@ -924,23 +854,13 @@ pgrecovery_loop(int fd)
 					status = send_packet(&Target,&packet);
 				}
 				/*
-				 * stop queueing, and re-initialize recovery status
+				 * stop queueing, and initiarise recovery status
 				 */
 				finish_recovery();
 				loop_end = true;
 				/*
 				 * send finish recovery to load balancer
 				 */
-				if (Master.recovery_sock != -1)
-				{
-					close(Master.recovery_sock);
-					Master.recovery_sock = -1;
-				}
-				if (Target.recovery_sock != -1)
-				{
-					close(Target.recovery_sock);
-					Target.recovery_sock = -1;
-				}
 				send_packet(&Master, &packet);
 				MasterPacketData.packet_no = packet.packet_no;
 				PGRsend_load_balance_packet(&MasterPacketData);
@@ -961,10 +881,6 @@ pgrecovery_loop(int fd)
 				PGRsend_load_balance_packet(&MasterPacketData);
 				memset((char *)&MasterPacketData,0,sizeof(RecoveryPacket));
 				break;
-		default:
-		  show_error("%s:unknown packet. abort to parse");
-			     loop_end=true;
-			     break;
 		}
 		if (loop_end)
 		{
@@ -990,9 +906,6 @@ PGRrecovery_main(void)
 	int fd = -1;
 	int rtn;
 
-#ifdef PRINT_DEBUG
-	show_debug("%s:PGRrecovery_main bind port %d",func,Recovery_Port_Number);
-#endif			
 	status = PGR_Create_Socket_Bind(&fd, "", Recovery_Port_Number);
 	if (status != STATUS_OK)
 	{
@@ -1007,6 +920,12 @@ PGRrecovery_main(void)
 		fd_set	  rmask;
 		struct timeval timeout;
 
+		Idle_Flag = IDLE_MODE;
+		if (Exit_Request)
+		{
+			exit(0);
+		}
+
 		timeout.tv_sec = RECOVERY_TIMEOUT;
 		timeout.tv_usec = 0;
 
@@ -1018,22 +937,157 @@ PGRrecovery_main(void)
 		rtn = select(fd+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
 		if (rtn && FD_ISSET(fd, &rmask))
 		{
+			Idle_Flag = BUSY_MODE;
 			pgrecovery_loop(fd);
 		}
 	}
 }
 
-#ifdef PRINT_DEBUG
 static void
 show_recovery_packet(RecoveryPacket * packet)
 {
-	show_debug("no = %d",ntohs(packet->packet_no));
-	show_debug("max_connect = %d",ntohs(packet->max_connect));
-	show_debug("port = %d",ntohs(packet->port));
-	show_debug("recoveryPort = %d",ntohs(packet->recoveryPort));
+	char * func = "show_recovery_packet()";
+
+#ifdef PRINT_DEBUG
+	show_debug("%s:no = %d",func,ntohs(packet->packet_no));
+	show_debug("%s:max_connect = %d",func,ntohs(packet->max_connect));
+	show_debug("%s:port = %d",func,ntohs(packet->port));
+	show_debug("%s:recoveryPort = %d",func,ntohs(packet->recoveryPort));
 	if (packet->hostName != NULL)
-		show_debug("hostName = %s",packet->hostName);
+		show_debug("%s:hostName = %s",func,packet->hostName);
 	if (packet->pg_data != NULL)
-		show_debug("pg_data = %s",packet->pg_data);
-}
+		show_debug("%s:pg_data = %s",func,packet->pg_data);
 #endif			
+}
+
+FILE *
+PGRget_recovery_queue_file_for_write(void)
+{
+	char * func = "PGRget_recovery_queue_file_for_write()";
+	int write_queue_no = 0;
+	int read_queue_no = 0;
+	char fname[256];
+
+	PGRsem_lock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+	/*
+	 * open recovery queue file
+	 */
+	write_queue_no = Recovery_Status_Inf->write_queue_no;
+	read_queue_no = Recovery_Status_Inf->read_queue_no;
+	if (write_queue_no == read_queue_no)
+	{
+		if (RecoveryQueue.queue_fp != (FILE *)NULL)
+		{
+			fflush(RecoveryQueue.queue_fp);
+			fclose(RecoveryQueue.queue_fp);
+			RecoveryQueue.queue_fp = (FILE *)NULL;
+		}
+		write_queue_no ++;
+		Recovery_Status_Inf->write_queue_no = write_queue_no;
+	}
+	if (write_queue_no != RecoveryQueue.current_queue_no)
+	{
+		if (RecoveryQueue.queue_fp != (FILE *)NULL)
+		{
+			fflush(RecoveryQueue.queue_fp);
+			fclose(RecoveryQueue.queue_fp);
+			RecoveryQueue.queue_fp = (FILE *)NULL;
+		}
+		RecoveryQueue.current_queue_no = write_queue_no;
+	}
+	if (RecoveryQueue.queue_fp == (FILE *)NULL)
+	{
+		snprintf(fname, sizeof(fname), "%s/%s%d", PGR_Write_Path, RECOVERY_QUEUE_FILE,write_queue_no);
+		RecoveryQueue.queue_fp = fopen(fname, "w+");
+		if (!RecoveryQueue.queue_fp)
+		{
+			show_error("%s:could not open recovery queue file as %s. reason: %s",
+					   func,fname, strerror(errno));
+			PGRsem_unlock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+			return (FILE*)NULL;
+		}
+	}
+	PGRsem_unlock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+
+	return RecoveryQueue.queue_fp;
+}
+
+FILE *
+PGRget_recovery_queue_file_for_read(int next)
+{
+	char * func = "PGRget_recovery_queue_file_for_read()";
+	int write_queue_no = 0;
+	int read_queue_no = 0;
+	char fname[256];
+
+	PGRsem_lock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+	/*
+	 * open recovery queue file
+	 */
+	read_queue_no = Recovery_Status_Inf->read_queue_no;
+	if (read_queue_no == 0)
+	{
+		read_queue_no ++;
+		Recovery_Status_Inf->read_queue_no = read_queue_no;
+	}
+	if (next)
+	{
+		if (read_queue_no < write_queue_no)
+		{
+			if (RecoveryQueue.queue_fp != NULL)
+			{
+				fflush(RecoveryQueue.queue_fp);
+				fclose(RecoveryQueue.queue_fp);
+				snprintf(fname, sizeof(fname), "%s/%s%d", PGR_Write_Path, RECOVERY_QUEUE_FILE,read_queue_no);
+				unlink(fname);
+				RecoveryQueue.queue_fp = NULL;
+			}
+			read_queue_no ++;
+			Recovery_Status_Inf->read_queue_no = read_queue_no;
+		}
+	}
+
+	if (RecoveryQueue.queue_fp == NULL)
+	{
+		snprintf(fname, sizeof(fname), "%s/%s%d", PGR_Write_Path, RECOVERY_QUEUE_FILE,read_queue_no);
+		RecoveryQueue.queue_fp = fopen(fname, "r");
+		if (!RecoveryQueue.queue_fp)
+		{
+			show_error("%s:could not open recovery queue file as %s. reason: %s",
+					   func,fname, strerror(errno));
+			PGRsem_unlock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+			return (FILE*)NULL;
+		}
+	}
+	PGRsem_unlock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+
+	return RecoveryQueue.queue_fp;
+}
+
+static void
+clear_recovery_queue_file(void)
+{
+	int write_queue_no = 0;
+	int read_queue_no = 0;
+	char fname[256];
+
+	PGRsem_lock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+	write_queue_no = Recovery_Status_Inf->write_queue_no;
+	read_queue_no = Recovery_Status_Inf->read_queue_no;
+	if (write_queue_no != read_queue_no)
+	{
+		return;
+	}
+	if (RecoveryQueue.queue_fp != NULL)
+	{
+		fflush(RecoveryQueue.queue_fp);
+		fclose(RecoveryQueue.queue_fp);
+		RecoveryQueue.queue_fp = (FILE *)NULL;
+	}
+	RecoveryQueue.current_queue_no = 0;
+	snprintf(fname, sizeof(fname), "%s/%s%d", PGR_Write_Path, RECOVERY_QUEUE_FILE,read_queue_no);
+	unlink(fname);
+	PGRsem_unlock(RecoverySemID, SEM_NUM_OF_RECOVERY_QUEUE);
+}
+
+#endif /* USE_REPLICATION */
